from typing import Optional, TypeVar, Generic, Tuple, Callable, Coroutine
from dataclasses import dataclass
from enum import Enum
from asyncio import Queue


CommonCallback = Optional[Callable[[dict], Coroutine[None, None, None]]]
T = TypeVar("T")


@dataclass
class GeoInfo:
    ip: str
    timezone: str
    geo: dict[str, float]
    country: str
    languages: list[str]
    city: str
    state_province: str
    proxy_type: str


class __Base(Generic[T]):
    def __init__(self, que: Queue[T]) -> None:
        self.que = que


class Sender(__Base[T]):
    async def send(self, data: T) -> None:
        await self.que.put(data)


class Receiver(__Base[T]):
    async def recv(self) -> T:
        return await self.que.get()


class Channel(Generic[T]):
    @classmethod
    def one_way(cls) -> Tuple[Sender[T], Receiver[T]]:
        que: Queue = Queue()
        return Sender[T](que), Receiver[T](que)


class DomainEvent(Enum):
    pass


@dataclass
class TargetConnectionInfo:
    description: str; devtoolsFrontendUrl: str
    id: str; title: str; type: str; url: str; webSocketDebuggerUrl: str
    faviconUrl: str = None; parentId: str = None


class TargetConnectionType(Enum):
    page = "page"
    background_page = "background_page"
    worker = "worker"
    service_worker = "service_worker"
    iframe ="iframe"
    other = "other"


@dataclass
class ViewportRect:
    """ Ширина и высота вьюпорта """
    width: int; height: int


@dataclass
class WindowRect:
    """ Ширина и высота окна браузера(outer - свойства) """
    width: int; height: int


class KeyModifiers(Enum):
    """ Клавиши-модификаторы """
    none = 0; alt = 1; ctrl = 2; meta = command = 4; shift = 8


@dataclass
class KeyEvents:
    """ https://gist.github.com/jhincapie/8a4c95d5cbe81d79b329ffc37e9f6c97 """
    backspace =           {"code": "Backspace",       "text": "",   "keyIdentifier": "U+0008",  "key": "Backspace", "windowsVirtualKeyCode": 8,   "nativeVirtualKeyCode": 8}
    tab =                 {"code": "Tab",             "text": "",   "keyIdentifier": "U+0009",  "key": "Tab",       "windowsVirtualKeyCode": 9,   "nativeVirtualKeyCode": 9}
    enter =               {"code": "Enter",           "text": "",   "keyIdentifier": "U+000D",  "key": "Enter",     "windowsVirtualKeyCode": 13,  "nativeVirtualKeyCode": 13}
    shift =               {"code": "Shift",           "text": "",   "keyIdentifier": "U+0010",  "key": "Shift",     "windowsVirtualKeyCode": 16,  "nativeVirtualKeyCode": 16}
    shift_left =          {"code": "ShiftLeft",       "text": "",   "keyIdentifier": "U+00A0",  "key": "Shift",     "windowsVirtualKeyCode": 160, "nativeVirtualKeyCode": 160}
    shift_right =         {"code": "ShiftRight",      "text": "",   "keyIdentifier": "U+00A1",  "key": "Shift",     "windowsVirtualKeyCode": 161, "nativeVirtualKeyCode": 161}
    control =             {"code": "Control",         "text": "",   "keyIdentifier": "Control", "key": "Control",   "windowsVirtualKeyCode": 17,  "nativeVirtualKeyCode": 17}
    control_left =        {"code": "ControlLeft",     "text": "",   "keyIdentifier": "U+00A2",  "key": "Control",   "windowsVirtualKeyCode": 162, "nativeVirtualKeyCode": 162}
    control_right =       {"code": "ControlRight",    "text": "",   "keyIdentifier": "U+00A3",  "key": "Control",   "windowsVirtualKeyCode": 163, "nativeVirtualKeyCode": 163}
    alt =                 {"code": "Alt",             "text": "",   "keyIdentifier": "Alt",     "key": "Alt",       "windowsVirtualKeyCode": 18,  "nativeVirtualKeyCode": 18}
    alt_left =            {"code": "AltLeft",         "text": "",   "keyIdentifier": "U+00A4",  "key": "Alt",       "windowsVirtualKeyCode": 164, "nativeVirtualKeyCode": 164}
    alt_right =           {"code": "AltRight",        "text": "",   "keyIdentifier": "U+00A5",  "key": "Alt",       "windowsVirtualKeyCode": 165, "nativeVirtualKeyCode": 165}
    escape =              {"code": "Escape",          "text": "",   "keyIdentifier": "U+001B",  "key": "Escape",    "windowsVirtualKeyCode": 27,  "nativeVirtualKeyCode": 27}
    space =               {"code": "Space",           "text": " ",  "keyIdentifier": "U+0020",  "key": " ",         "windowsVirtualKeyCode": 32,  "nativeVirtualKeyCode": 32}
    arrow_left =          {"code": "ArrowLeft",       "text": "",   "keyIdentifier": "U+0025",  "key": "ArrowLeft", "windowsVirtualKeyCode": 37,  "nativeVirtualKeyCode": 37}
    arrow_up =            {"code": "ArrowUp",         "text": "",   "keyIdentifier": "U+0026",  "key": "ArrowUp",   "windowsVirtualKeyCode": 38,  "nativeVirtualKeyCode": 38}
    arrow_right =         {"code": "ArrowRight",      "text": "",   "keyIdentifier": "U+0027",  "key": "ArrowRight","windowsVirtualKeyCode": 39,  "nativeVirtualKeyCode": 39}
    arrow_down =          {"code": "ArrowDown",       "text": "",   "keyIdentifier": "U+0028",  "key": "ArrowDown", "windowsVirtualKeyCode": 40,  "nativeVirtualKeyCode": 40}
    parentheses_right =   {"code": "Parentheses",     "text": ")",  "keyIdentifier": "U+0030",  "key": ")",         "windowsVirtualKeyCode": 48,  "nativeVirtualKeyCode": 48}
    digit0 =              {"code": "Digit0",          "text": "0",  "keyIdentifier": "U+0030",  "key": "0",         "windowsVirtualKeyCode": 48,  "nativeVirtualKeyCode": 48}
    digit1 =              {"code": "Digit1",          "text": "1",  "keyIdentifier": "U+0031",  "key": "1",         "windowsVirtualKeyCode": 49,  "nativeVirtualKeyCode": 49}
    digit2 =              {"code": "Digit2",          "text": "2",  "keyIdentifier": "U+0032",  "key": "2",         "windowsVirtualKeyCode": 50,  "nativeVirtualKeyCode": 50}
    at =                  {"code": "At",              "text": "@",  "keyIdentifier": "U+0032",  "key": "@",         "windowsVirtualKeyCode": 50,  "nativeVirtualKeyCode": 50}
    digit3 =              {"code": "Digit3",          "text": "3",  "keyIdentifier": "U+0033",  "key": "3",         "windowsVirtualKeyCode": 51,  "nativeVirtualKeyCode": 51}
    hash =                {"code": "Hash",            "text": "#",  "keyIdentifier": "U+0033",  "key": "#",         "windowsVirtualKeyCode": 51,  "nativeVirtualKeyCode": 51}   # #
    number =              {"code": "Number",          "text": "№",  "keyIdentifier": "U+0033",  "key": "№",         "windowsVirtualKeyCode": 51,  "nativeVirtualKeyCode": 51}   # №
    digit4 =              {"code": "Digit4",          "text": "4",  "keyIdentifier": "U+0034",  "key": "4",         "windowsVirtualKeyCode": 52,  "nativeVirtualKeyCode": 52}
    dollar =              {"code": "Dollar",          "text": "$",  "keyIdentifier": "U+0034",  "key": "$",         "windowsVirtualKeyCode": 52,  "nativeVirtualKeyCode": 52}   # $
    digit5 =              {"code": "Digit5",          "text": "5",  "keyIdentifier": "U+0035",  "key": "5",         "windowsVirtualKeyCode": 53,  "nativeVirtualKeyCode": 53}
    percent =             {"code": "Percent",         "text": "%",  "keyIdentifier": "U+0035",  "key": "%",         "windowsVirtualKeyCode": 53,  "nativeVirtualKeyCode": 53}   # %
    digit6 =              {"code": "Digit6",          "text": "6",  "keyIdentifier": "U+0036",  "key": "6",         "windowsVirtualKeyCode": 54,  "nativeVirtualKeyCode": 54}
    angle_up =            {"code": "AngleUp",         "text": "^",  "keyIdentifier": "U+0036",  "key": "^",         "windowsVirtualKeyCode": 54,  "nativeVirtualKeyCode": 54}   # ^
    digit7 =              {"code": "Digit7",          "text": "7",  "keyIdentifier": "U+0037",  "key": "7",         "windowsVirtualKeyCode": 55,  "nativeVirtualKeyCode": 55}
    ampersand =           {"code": "Ampersand",       "text": "&",  "keyIdentifier": "U+0037",  "key": "&",         "windowsVirtualKeyCode": 55,  "nativeVirtualKeyCode": 55}   # &
    digit8 =              {"code": "Digit8",          "text": "8",  "keyIdentifier": "U+0038",  "key": "8",         "windowsVirtualKeyCode": 56,  "nativeVirtualKeyCode": 56}
    digit9 =              {"code": "Digit9",          "text": "9",  "keyIdentifier": "U+0039",  "key": "9",         "windowsVirtualKeyCode": 57,  "nativeVirtualKeyCode": 57}
    parentheses_left =    {"code": "Parentheses",     "text": "(",  "keyIdentifier": "U+0039",  "key": "(",         "windowsVirtualKeyCode": 57,  "nativeVirtualKeyCode": 57}
    keyA =                {"code": "KeyA",            "text": "a",  "keyIdentifier": "U+0041",  "key": "a",         "windowsVirtualKeyCode": 65,  "nativeVirtualKeyCode": 65}
    keyB =                {"code": "KeyB",            "text": "b",  "keyIdentifier": "U+0042",  "key": "b",         "windowsVirtualKeyCode": 66,  "nativeVirtualKeyCode": 66}
    keyC =                {"code": "KeyC",            "text": "c",  "keyIdentifier": "U+0043",  "key": "c",         "windowsVirtualKeyCode": 67,  "nativeVirtualKeyCode": 67}
    keyD =                {"code": "KeyD",            "text": "d",  "keyIdentifier": "U+0044",  "key": "d",         "windowsVirtualKeyCode": 68,  "nativeVirtualKeyCode": 68}
    keyE =                {"code": "KeyE",            "text": "e",  "keyIdentifier": "U+0045",  "key": "e",         "windowsVirtualKeyCode": 69,  "nativeVirtualKeyCode": 69}
    keyF =                {"code": "KeyF",            "text": "f",  "keyIdentifier": "U+0046",  "key": "f",         "windowsVirtualKeyCode": 70,  "nativeVirtualKeyCode": 70}
    keyG =                {"code": "KeyG",            "text": "g",  "keyIdentifier": "U+0047",  "key": "g",         "windowsVirtualKeyCode": 71,  "nativeVirtualKeyCode": 71}
    keyH =                {"code": "KeyH",            "text": "h",  "keyIdentifier": "U+0048",  "key": "h",         "windowsVirtualKeyCode": 72,  "nativeVirtualKeyCode": 72}
    keyI =                {"code": "KeyI",            "text": "i",  "keyIdentifier": "U+0049",  "key": "i",         "windowsVirtualKeyCode": 73,  "nativeVirtualKeyCode": 73}
    keyJ =                {"code": "KeyJ",            "text": "j",  "keyIdentifier": "U+004A",  "key": "j",         "windowsVirtualKeyCode": 74,  "nativeVirtualKeyCode": 74}
    keyK =                {"code": "KeyK",            "text": "k",  "keyIdentifier": "U+004B",  "key": "k",         "windowsVirtualKeyCode": 75,  "nativeVirtualKeyCode": 75}
    keyL =                {"code": "KeyL",            "text": "l",  "keyIdentifier": "U+004C",  "key": "l",         "windowsVirtualKeyCode": 76,  "nativeVirtualKeyCode": 76}
    keyM =                {"code": "KeyM",            "text": "m",  "keyIdentifier": "U+004D",  "key": "m",         "windowsVirtualKeyCode": 77,  "nativeVirtualKeyCode": 77}
    keyN =                {"code": "KeyN",            "text": "n",  "keyIdentifier": "U+004E",  "key": "n",         "windowsVirtualKeyCode": 78,  "nativeVirtualKeyCode": 78}
    keyO =                {"code": "KeyO",            "text": "o",  "keyIdentifier": "U+004F",  "key": "o",         "windowsVirtualKeyCode": 79,  "nativeVirtualKeyCode": 79}
    keyP =                {"code": "KeyP",            "text": "p",  "keyIdentifier": "U+0050",  "key": "p",         "windowsVirtualKeyCode": 80,  "nativeVirtualKeyCode": 80}
    keyQ =                {"code": "KeyQ",            "text": "q",  "keyIdentifier": "U+0051",  "key": "q",         "windowsVirtualKeyCode": 81,  "nativeVirtualKeyCode": 81}
    keyR =                {"code": "KeyR",            "text": "r",  "keyIdentifier": "U+0052",  "key": "r",         "windowsVirtualKeyCode": 82,  "nativeVirtualKeyCode": 82}
    keyS =                {"code": "KeyS",            "text": "s",  "keyIdentifier": "U+0053",  "key": "s",         "windowsVirtualKeyCode": 83,  "nativeVirtualKeyCode": 83}
    keyT =                {"code": "KeyT",            "text": "t",  "keyIdentifier": "U+0054",  "key": "t",         "windowsVirtualKeyCode": 84,  "nativeVirtualKeyCode": 84}
    keyU =                {"code": "KeyU",            "text": "u",  "keyIdentifier": "U+0055",  "key": "u",         "windowsVirtualKeyCode": 85,  "nativeVirtualKeyCode": 85}
    keyV =                {"code": "KeyV",            "text": "v",  "keyIdentifier": "U+0056",  "key": "v",         "windowsVirtualKeyCode": 86,  "nativeVirtualKeyCode": 86}
    keyW =                {"code": "KeyW",            "text": "w",  "keyIdentifier": "U+0057",  "key": "w",         "windowsVirtualKeyCode": 87,  "nativeVirtualKeyCode": 87}
    keyX =                {"code": "KeyX",            "text": "x",  "keyIdentifier": "U+0058",  "key": "x",         "windowsVirtualKeyCode": 88,  "nativeVirtualKeyCode": 88}
    keyY =                {"code": "KeyY",            "text": "y",  "keyIdentifier": "U+0059",  "key": "y",         "windowsVirtualKeyCode": 89,  "nativeVirtualKeyCode": 89}
    keyZ =                {"code": "KeyZ",            "text": "z",  "keyIdentifier": "U+005A",  "key": "z",         "windowsVirtualKeyCode": 90,  "nativeVirtualKeyCode": 90}
    metaLeft =            {"code": "MetaLeft",        "text": "",   "keyIdentifier": "U+005B",  "key": "Meta",      "windowsVirtualKeyCode": 91,  "nativeVirtualKeyCode": 91}
    metaRight =           {"code": "MetaRight",       "text": "",   "keyIdentifier": "U+005D",  "key": "Meta",      "windowsVirtualKeyCode": 93,  "nativeVirtualKeyCode": 93}
    context_menu =        {"code": "ContextMenu",     "text": "",   "keyIdentifier": "U+005D",  "key": "",          "windowsVirtualKeyCode": 93,  "nativeVirtualKeyCode": 93}
    equal =               {"code": "Equal",           "text": "=",  "keyIdentifier": "U+003D",  "key": "=",         "windowsVirtualKeyCode": 61,  "nativeVirtualKeyCode": 61}
    multiply =            {"code": "Multipy",         "text": "*",  "keyIdentifier": "U+006A",  "key": "*",         "windowsVirtualKeyCode": 106, "nativeVirtualKeyCode": 106}   # Numpad *
    oem_plus =            {"code": "Plus",            "text": "+",  "keyIdentifier": "U+00BB",  "key": "+",         "windowsVirtualKeyCode": 187, "nativeVirtualKeyCode": 187}   # OEM_PLUS(+ =)
    numpad_plus =         {"code": "Plus",            "text": "+",  "keyIdentifier": "U+006B",  "key": "+",         "windowsVirtualKeyCode": 107, "nativeVirtualKeyCode": 107}   #
    oem_minus =           {"code": "Minus",           "text": "-",  "keyIdentifier": "U+00BD",  "key": "-",         "windowsVirtualKeyCode": 189, "nativeVirtualKeyCode": 189}   # OEM_MINUS(_ -)
    underscore =          {"code": "Underscore",      "text": "_",  "keyIdentifier": "U+00BD",  "key": "_",         "windowsVirtualKeyCode": 189, "nativeVirtualKeyCode": 189}   # OEM_MINUS(_ -)
    numpad_minus =        {"code": "Minus",           "text": "-",  "keyIdentifier": "U+006D",  "key": "-",         "windowsVirtualKeyCode": 109, "nativeVirtualKeyCode": 109}   #
    bracket_left =        {"code": "BracketLeft",     "text": "[",  "keyIdentifier": "U+005B",  "key": "[",         "windowsVirtualKeyCode": 219, "nativeVirtualKeyCode": 219}
    bracket_right =       {"code": "BracketRight",    "text": "]",  "keyIdentifier": "U+005D",  "key": "]",         "windowsVirtualKeyCode": 221, "nativeVirtualKeyCode": 221}
    brace_left =          {"code": "BraceLeft",       "text": "{",  "keyIdentifier": "U+00DB",  "key": "{",         "windowsVirtualKeyCode": 219, "nativeVirtualKeyCode": 219}
    brace_right =         {"code": "BraceRight",      "text": "}",  "keyIdentifier": "U+00DD",  "key": "}",         "windowsVirtualKeyCode": 221, "nativeVirtualKeyCode": 221}
    pipe =                {"code": "Pipe",            "text": "|",  "keyIdentifier": "U+00DC",  "key": "|",         "windowsVirtualKeyCode": 220, "nativeVirtualKeyCode": 220}
    backslash =           {"code": "Backslash",       "text": "\\", "keyIdentifier": "U+00DC",  "key": "\\",        "windowsVirtualKeyCode": 220, "nativeVirtualKeyCode": 220}
    quote =               {"code": "Quote",           "text": "'",  "keyIdentifier": "U+00DE",  "key": "'",         "windowsVirtualKeyCode": 222, "nativeVirtualKeyCode": 222}
    double_quote =        {"code": "DoubleQuote",     "text": "\"", "keyIdentifier": "U+00DE",  "key": "\"",        "windowsVirtualKeyCode": 222, "nativeVirtualKeyCode": 222}
    exclamation_mark =    {"code": "ExclamationMark", "text": "!",  "keyIdentifier": "U+00DF",  "key": "!",         "windowsVirtualKeyCode": 223, "nativeVirtualKeyCode": 223}  # !
    semicolon =           {"code": "Semicolon",       "text": ";",  "keyIdentifier": "U+00BA",  "key": ";",         "windowsVirtualKeyCode": 186, "nativeVirtualKeyCode": 186}
    colon =               {"code": "Colon",           "text": ":",  "keyIdentifier": "U+00BA",  "key": ":",         "windowsVirtualKeyCode": 186, "nativeVirtualKeyCode": 186}
    comma =               {"code": "Comma",           "text": ",",  "keyIdentifier": "U+002C",  "key": ",",         "windowsVirtualKeyCode": 188, "nativeVirtualKeyCode": 188}
    angle_bracket_left =  {"code": "AngleBracket",    "text": "<",  "keyIdentifier": "U+002C",  "key": "<",         "windowsVirtualKeyCode": 188, "nativeVirtualKeyCode": 188}  # <
    dot =                 {"code": "Dot",             "text": ".",  "keyIdentifier": "U+002E",  "key": ".",         "windowsVirtualKeyCode": 190, "nativeVirtualKeyCode": 190}
    period =              {"code": "Period",          "text": ".",  "keyIdentifier": "U+002E",  "key": ".",         "windowsVirtualKeyCode": 190, "nativeVirtualKeyCode": 190}
    angle_bracket_right = {"code": "AngleBracket",    "text": ">",  "keyIdentifier": "U+002E",  "key": ">",         "windowsVirtualKeyCode": 190, "nativeVirtualKeyCode": 190}  # >
    slash =               {"code": "Slash",           "text": "/",  "keyIdentifier": "U+00BF",  "key": "/",         "windowsVirtualKeyCode": 191, "nativeVirtualKeyCode": 191}
    question_mark =       {"code": "QuestionMark",    "text": "?",  "keyIdentifier": "U+00BF",  "key": "?",         "windowsVirtualKeyCode": 191, "nativeVirtualKeyCode": 191}  # ?
    backquote =           {"code": "Backquote",       "text": "`",  "keyIdentifier": "U+0060",  "key": "`",         "windowsVirtualKeyCode": 192, "nativeVirtualKeyCode": 192}
    tilda =               {"code": "Tilda",           "text": "~",  "keyIdentifier": "U+00C0",  "key": "~",         "windowsVirtualKeyCode": 192, "nativeVirtualKeyCode": 192}



WINDOWS_KEY_SET = {
    'LBUTTON': 1,   # Left mouse button
    'RBUTTON': 2,   # Right mouse button
    'CANCEL': 3,    # Control-break processing
    'MBUTTON': 4,   # Middle mouse button (three-button mouse)
    'XBUTTON1': 5,  # X1 mouse button
    'XBUTTON2': 6,  # X2 mouse button

    'BACK': 8,      # BACKSPACE key
    'BACKSPACE': 8, # BACKSPACE key
    'TAB': 9,       # TAB key
    'CLEAR': 12,    # CLEAR key
    'ENTER': 13,    # ENTER key
    'SHIFT': 16,    # SHIFT key
    'CONTROL': 17,  # CTRL key
    'ALT': 18,      # ALT key
    'PAUSE': 19,    # PAUSE key
    'CAPS': 20,     # CAPS LOCK key

    'KANA': 21,     # IME Kana mode
    'HANGUEL': 21,  # IME Hanguel mode (maintained for compatibility; use VK_HANGUL)
    'HANGUL': 21,   # IME Hangul mode
    'JUNJA': 23,    # IME Junja mode
    'FINAL': 24,    # IME final mode
    'HANJA': 25,    # IME Hanja mode
    'KANJI': 25,    # IME Kanji mode

    'CONVERT': 28,     # IME convert
    'NONCONVERT': 29,  # IME nonconvert
    'ACCEPT': 30,      # IME accept
    'MODECHANGE': 31,  # IME mode change request

    'ESCAPE': 27,    # ESC key
    'ESC': 27,       # ESC key
    'SPACE': 32,     # Пробел
    'PRIOR': 33,     # PAGE UP key
    'NEXT': 34,      # PAGE DOWN key
    'END': 35,       # END key
    'HOME': 36,      # HOME key
    'LEFT': 37,      # LEFT ARROW key
    'UP': 38,        # UP ARROW key
    'RIGHT': 39,     # RIGHT ARROW key
    'DOWN': 40,      # DOWN ARROW key
    'SELECT': 41,    # SELECT key
    'PRINT': 42,     # PRINT key
    'EXECUTE': 43,   # EXECUTE key
    'SNAPSHOT': 44,  # PRINT SCREEN key
    'INSERT': 45,    # INS key
    'DELETE': 46,    # DEL key
    'DEL': 46,       # DEL key
    'HELP': 47,      # HELP key

    '0': 48,  # 0 key
    '1': 49,  # 1 key
    '2': 50,  # 2 key
    '3': 51,  # 3 key
    '4': 52,  # 4 key
    '5': 53,  # 5 key
    '6': 54,  # 6 key
    '7': 55,  # 7 key
    '8': 56,  # 8 key
    '9': 57,  # 9 key
    'A': 65,
    'B': 66,
    'C': 67,
    'D': 68,
    'E': 69,
    'F': 70,
    'G': 71,
    'H': 72,
    'I': 73,
    'J': 74,
    'K': 75,
    'L': 76,
    'M': 77,
    'N': 78,
    'O': 79,
    'P': 80,
    'Q': 81,
    'R': 82,
    'S': 83,
    'T': 84,
    'U': 85,
    'V': 86,
    'W': 87,
    'X': 88,
    'Y': 89,
    'Z': 90,

    'LWIN': 91,        # Left Windows key (Natural keyboard)
    'RWIN': 92,        # Right Windows key (Natural keyboard)
    'APPS': 93,        # Applications key (Natural keyboard)
    'SLEEP': 95,       # Computer Sleep key
    'NUMPAD0': 96,     # Numeric keypad 0 key
    'NUMPAD1': 97,     # Numeric keypad 1 key
    'NUMPAD2': 98,     # Numeric keypad 2 key
    'NUMPAD3': 99,     # Numeric keypad 3 key
    'NUMPAD4': 100,    # Numeric keypad 4 key
    'NUMPAD5': 101,    # Numeric keypad 5 key
    'NUMPAD6': 102,    # Numeric keypad 6 key
    'NUMPAD7': 103,    # Numeric keypad 7 key
    'NUMPAD8': 104,    # Numeric keypad 8 key
    'NUMPAD9': 105,    # Numeric keypad 9 key
    'MULTIPLY': 106,   # Multiply key
    'ADD': 107,        # Add key
    'SEPARATOR': 108,  # Separator key
    'SUBTRACT': 109,   # Subtract key
    'DECIMAL': 110,    # Decimal key
    'DIVIDE': 111,     # Divide key

    'F1':  112,  # F1 key
    'F2':  113,  # F2 key
    'F3':  114,  # F3 key
    'F4':  115,  # F4 key
    'F5':  116,  # F5 key
    'F6':  117,  # F6 key
    'F7':  118,  # F7 key
    'F8':  119,  # F8 key
    'F9':  120,  # F9 key
    'F10': 121,  # F10 key
    'F11': 122,  # F11 key
    'F12': 123,  # F12 key
    'F13': 124,  # F13 key
    'F14': 125,  # F14 key
    'F15': 126,  # F15 key
    'F16': 127,  # F16 key
    'F17': 128,  # F17 key
    'F18': 129,  # F18 key
    'F19': 130,  # F19 key
    'F20': 131,  # F20 key
    'F21': 132,  # F21 key
    'F22': 133,  # F22 key
    'F23': 134,  # F23 key
    'F24': 135,  # F24 key

    'NUMLOCK': 144,              #NUM LOCK key
    'SCROLL': 145,               #SCROLL LOCK key
    'LSHIFT': 160,               #Left SHIFT key
    'RSHIFT': 161,               #Right SHIFT key
    'LCONTROL': 162,             #Left CONTROL key
    'RCONTROL': 163,             #Right CONTROL key
    'LMENU': 164,                #Left MENU key
    'RMENU': 165,                #Right MENU key
    'BROWSER_BACK': 166,         #Browser Back key
    'BROWSER_FORWARD': 167,      #Browser Forward key
    'BROWSER_REFRESH': 168,      #Browser Refresh key
    'BROWSER_STOP': 169,         #Browser Stop key
    'BROWSER_SEARCH': 170,       #Browser Search key
    'BROWSER_FAVORITES': 171,    #Browser Favorites key
    'BROWSER_HOME': 172,         #Browser Start and Home key
    'VOLUME_MUTE': 173,          #Volume Mute key
    'VOLUME_DOWN': 174,          #Volume Down key
    'VOLUME_UP': 175,            #Volume Up key
    'MEDIA_NEXT_TRACK': 176,     #Next Track key
    'MEDIA_PREV_TRACK': 177,     #Previous Track key
    'MEDIA_STOP': 178,           #Stop Media key
    'MEDIA_PLAY_PAUSE': 179,     #Play/Pause Media key
    'LAUNCH_MAIL': 180,          #Start Mail key
    'LAUNCH_MEDIA_SELECT': 181,  #Select Media key
    'LAUNCH_APP1': 182,          #Start Application 1 key
    'LAUNCH_APP2': 183,          #Start Application 2 key

    'OEM_1': 186,       #Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the ';:' key
    'OEM_PLUS': 187,    #For any country/region, the '+' key
    'OEM_COMMA': 188,   #For any country/region, the ',' key
    'OEM_MINUS': 189,   #For any country/region, the '-' key
    'OEM_PERIOD': 190,  #For any country/region, the '.' key
    'OEM_2': 191,       #Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '/?' key =
    'OEM_3': 192,       #Used for miscellaneous characters; it can vary by keyboard. = For the US standard keyboard, the '`~' key
    'OEM_4': 219,       #Used for miscellaneous characters; it can vary by keyboard. For the US standard keyboard, the '[{' key
    'OEM_5': 220,       #Used for miscellaneous characters; it can vary by keyboard.For the US standard keyboard, the '\|' key
    'OEM_6': 221,       #Used for miscellaneous characters; it can vary by keyboard.For the US standard keyboard, the ']}' key
    'OEM_7': 222,       #Used for miscellaneous characters; it can vary by keyboard.For the US standard keyboard, the 'single-quote/double-quote' key
    'OEM_8': 223,       #Used for miscellaneous characters; it can vary by keyboard.
    'OEM_102': 226,     #Either the angle bracket key or the backslash key on the RT 102-key keyboard 0xE3-E4 OEM specific

    'PROCESSKEY': 229,  #IME PROCESS key 0xE6 = OEM specific #
    'PACKET': 231,      #Used to pass Unicode characters as if they were keystrokes. The VK_PACKET key is the low word of a 32-bit Virtual "key" value used for non-keyboard input methods. For more information, see Remark in KEYBDINPUT, SendInput, WM_KEYDOWN, and WM_KEYUP #-
    'ATTN': 246,        #Attn key
    'CRSEL': 247,       #CrSel key
    'EXSEL': 248,       #ExSel key
    'EREOF': 249,       #Erase EOF key
    'PLAY': 250,        #Play key
    'ZOOM': 251,        #Zoom key
    'NONAME': 252,      #Reserved
    'PA1': 253,         #PA1 key
    'OEM_CLEAR': 254,   #
}



from typing import List
from ...data import DomainEvent


class CSS:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Log
    #   LogEntry -> https://chromedevtools.github.io/devtools-protocol/tot/Log#type-LogEntry
    """
    __slots__ = ("_connection", "enabled", "_style_sheets")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled = False
        self._style_sheets: List[str] = []  # Если домен CSS активирован, сюда попадут все 'styleSheetId' страницы

    @property
    def style_sheets(self) -> List[str]:
        return self._style_sheets

    @style_sheets.setter
    def style_sheets(self, value) -> None:
        self._style_sheets = value

    async def _CSS_sheet_catcher(self, data: dict) -> None:
        """
        Колбэк вызываемый для каждого 'CSS.styleSheetAdded'-события, если
            включён агент домена 'CSS'.
        """
        self._style_sheets.append(data["header"]["styleSheetId"])

    async def CSSEnable(self) -> None:
        """
        Включает агент CSS. Клиент не должен предполагать, что агент CSS включен,
            пока не будет получен результат этой команды.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-disable
        :return:
        """
        if not self.enabled:
            await self._connection.addListenerForEvent("CSS.styleSheetAdded", self._CSS_sheet_catcher)
            await self._connection.call("CSS.enable")
            self.enabled = True

    async def CSSDisable(self) -> None:
        """
        Отключает агент CSS.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-disable
        :return:
        """
        if self.enabled:
            self._connection.removeListenerForEvent("CSS.styleSheetAdded", self._CSS_sheet_catcher)
            await self._connection.call("CSS.disable")
            self.style_sheets = []
            self.enabled = False

    async def AddRule(
        self, styleSheetId: str, ruleText: str,
        startLine:   int = 0,
        startColumn: int = 0,
        endLine:     int = 0,
        endColumn:   int = 0
    ) -> dict:
        """
        Вставляет новое правило с заданным ruleText в таблицу стилей с заданным styleSheetId в позицию,
            указанную с помощью location.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-addRule
        :param styleSheetId:            Идентификатор стилей.
        :param ruleText:                Текст правила (например: 'body { background: red }').

        Текстовый диапазон в ресурсе. Все числа отсчитываются от нуля.
        :param startLine:               (optional) - Стартовая линия диапазона.
        :param startColumn:             (optional) - Начальный столбец диапазона (включительно).
        :param endLine:                 (optional) - Конечная строка диапазона.
        :param endColumn:               (optional) - Конечный столбец диапазона (исключая).
        :return:    https://chromedevtools.github.io/devtools-protocol/tot/CSS/#type-CSSRule
        """
        location = {"startLine": startLine, "startColumn": startColumn, "endLine": endLine, "endColumn": endColumn}
        args = {
            "styleSheetId": styleSheetId,
            "ruleText": ruleText,
            "location": location
        }
        return (await self._connection.call("CSS.addRule", args))["rule"]

    async def CollectClassNames(self, styleSheetId: str) -> List[str]:
        """
        Возвращает список всех имён классов для указанного ID.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-collectClassNames
        :param styleSheetId:            Идентификатор стилей.
        :return:
        """
        return (await self._connection.call("CSS.collectClassNames", {"styleSheetId": styleSheetId}))["classNames"]

    async def CreateStyleSheet(self, frameId: str = None) -> str:
        """
        Создает новую специальную таблицу стилей через "инспектор" во фрейме с заданным frameId.
            Если frameId не указан, будет использован frameId главного фрейма текущей страницы.
            Возвращает ID созданного styleSheet.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-createStyleSheet
        :param frameId:                 Идентификатор фрейма, в котором будет создан styleSheet.
        :return:                styleSheetId
        """
        frameId = frameId if frameId else self._connection.conn_id
        styleSheetId = (await self._connection.call("CSS.createStyleSheet", {"frameId": frameId}))["styleSheetId"]
        self._style_sheets.append(styleSheetId)
        return styleSheetId

    async def GetStyleSheetText(self, styleSheetId: str) -> str:
        """
        Возвращает текстовый контент стилей для указанного ID.
        https://chromedevtools.github.io/devtools-protocol/tot/CSS/#method-getStyleSheetText
        :param styleSheetId:            Идентификатор стилей.
        :return:
        """
        return (await self._connection.call("CSS.getStyleSheetText", {"styleSheetId": styleSheetId}))["text"]

class CSSEvent(DomainEvent):
    fontsUpdated = "CSS.fontsUpdated"
    mediaQueryResultChanged = "CSS.mediaQueryResultChanged"
    styleSheetAdded = "CSS.styleSheetAdded"
    styleSheetChanged = "CSS.styleSheetChanged"
    styleSheetRemoved = "CSS.styleSheetRemoved"



from typing import Optional, List, Callable, Awaitable
from ...data import DomainEvent
from .types import EventRequestPaused, EventAuthRequired, HeaderEntry, RequestPattern


class Fetch:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Fetch
    """
    __slots__ = ("_connection", "enabled")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled = False

    async def enable(
        self,
        patterns: Optional[List[RequestPattern]] = None,
        handleAuthRequests: bool = False,
        on_pause: Optional[Callable[[EventRequestPaused], Awaitable[None]]] = None,
        on_auth:  Optional[Callable[[EventAuthRequired], Awaitable[None]]] = None,
    ) -> None:
        """
        Включает выдачу событий requestPaused. Запрос будет приостановлен до тех пор,
            пока клиент не вызовет одну из функций failRequest, fulfillRequest или
            continueRequest / continueWithAuth.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-enable
        :param patterns:            (optional) Если указано, только запросы,
                                        соответствующие любому из этих шаблонов, будут
                                        вызывать событие fetchRequested и будут
                                        приостановлены до ответа клиента. Если не
                                        установлен, все запросы будут затронуты.
                                            https://chromedevtools.github.io/devtools-protocol/tot/Fetch#type-RequestPattern
        :param handleAuthRequests:  (optional) Если True - события authRequired будут
                                        выдаваться и запросы будут приостановлены в
                                        ожидании вызова continueWithAuth.
        :param on_pause:            (optional) Корутина, которая будет получать все события "requestPaused".
        :param on_auth:             (optional) Корутина, которая будет получать все события "authRequired".
        :return:
        """

        async def on_pause_decorator(params: dict) -> None:
            await on_pause(EventRequestPaused(**params))

        async def on_auth_decorator(params: dict) -> None:
            await on_auth(EventAuthRequired(**params))

        if on_pause is not None:
            await self._connection.addListenerForEvent(
                FetchEvent.requestPaused, on_pause_decorator)

        if on_auth is not None:
            await self._connection.addListenerForEvent(
                FetchEvent.authRequired, on_auth_decorator)

        if not self.enabled:
            args = {}
            patterns = patterns if patterns is not None else []
            if patterns:
                args.update({"patterns": [p.dict() for p in patterns]})
            if handleAuthRequests: args.update({"handleAuthRequests": handleAuthRequests})
            await self._connection.call("Fetch.enable", args)
            self.enabled = True

    async def disable(self) -> None:
        """
        Отключает взаимодействие с доменом.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-disable
        :return:
        """
        if self.enabled:
            await self._connection.call("Fetch.disable")
            self.enabled = False

    async def failRequest(self, requestId: str, errorReason: str) -> None:
        """
        Вызывает сбой запроса по указанной причине.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-failRequest
        :param requestId:           Идентификатор, полученный клиентом в событии requestPaused.
        :param errorReason:         Причина сбоя выборки на уровне сети. Возможные значения:
                                        Failed, Aborted, TimedOut, AccessDenied,
                                        ConnectionClosed, ConnectionReset,
                                        ConnectionRefused, ConnectionAborted,
                                        ConnectionFailed, NameNotResolved,
                                        InternetDisconnected, AddressUnreachable,
                                        BlockedByClient, BlockedByResponse
        :return:
        """
        await self._connection.call("Fetch.failRequest", {"requestId": requestId, "errorReason": errorReason})

    async def fulfillRequest(
            self, requestId: str,
            responseCode:                     int = 200,
            responseHeaders: Optional[List[HeaderEntry]] = None,
            binaryResponseHeaders:  Optional[str] = None,
            body:                   Optional[str] = None,
            responsePhrase:         Optional[str] = None,
            wait_for_response:               bool = False
    ) -> None:
        """
        Предоставляет браузеру ответ на запрос.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-fulfillRequest
        :param requestId:               Идентификатор, полученный клиентом в событии requestPaused.
        :param responseCode:            Код ответа HTTP(например - 200).
        :param responseHeaders:         (optional) Заголовки ответа. Например:
                                            [
                                                { "name": "User-Agent", "value": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36" },
                                                { "name": "Content-Type", "value": "application/json; charset=UTF-8" }
                                            ]
        :param binaryResponseHeaders:   (optional) Альтернативный способ указания заголовков ответа в
                                            виде разделенных \0 серий пар имя-значение. Описанный выше
                                            метод предпочтительней, если вам не нужно представлять
                                            некоторые значения, отличные от UTF8, которые не могут быть
                                            переданы по протоколу, в виде текста.
        :param body:                    (optional) Тело ответа, кодированное в строку формата base64.
        :param responsePhrase:          (optional) Текстовое представление responseCode. Если
                                            отсутствует, используется стандартная фраза,
                                            соответствующая responseCode.
        :param wait_for_response:       (optional) Дожидаться ответа?
        :return:
        """
        args = {"requestId": requestId, "responseCode": responseCode}
        if responseHeaders is not None:
            headers = [h.__dict__ for h in responseHeaders]
            args.update({"responseHeaders": headers})

        if binaryResponseHeaders is not None: args.update({"binaryResponseHeaders": binaryResponseHeaders})
        if body is not None: args.update({"body": body})
        if responsePhrase is not None: args.update({"responsePhrase": responsePhrase})
        # print("fulfillRequest args", json.dumps(args, indent=4))
        await self._connection.call("Fetch.fulfillRequest", args, wait_for_response=wait_for_response)

    async def continueRequest(
        self, requestId: str,
        url:                Optional[str] = None,
        method:             Optional[str] = None,
        postData:           Optional[str] = None,
        headers:     Optional[List[dict]] = None,
        interceptResponse: Optional[bool] = None,
        wait_for_response:           bool = False
    ) -> None:
        """
        Продолжает запрос, дополнительно изменяя некоторые его параметры.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-continueRequest
        :param requestId:           Идентификатор, полученный клиентом в событии requestPaused.
        :param url:                 (optional) Если установлено, URL-адрес запроса будет изменен так,
                                        чтобы страница не наблюдалась.
        :param method:              (optional) Переопределяет метод запроса переданным значением.
        :param postData:            (optional) Переопределяет данные запроса переданными.
        :param headers:             (optional) Переопределяет заголовки запроса переданными. . Например:
                                        [
                                            { "name": "User-Agent", "value": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.149 Safari/537.36" },
                                            { "name": "Content-Type", "value": "application/json; charset=UTF-8" }
                                        ]
        :param interceptResponse:   (optional) Если установлено, переопределяет поведение перехвата ответа для этого запроса.
        :param wait_for_response:   (optional) Дожидаться ответа?
        :return:
        """
        args = {"requestId": requestId}
        if url is not None: args.update({"url": url})
        if method is not None: args.update({"method": method})
        if postData is not None: args.update({"postData": postData})
        if headers is not None: args.update({"headers": headers})
        if interceptResponse is not None: args.update({"interceptResponse": interceptResponse})
        await self._connection.call("Fetch.continueRequest", args, wait_for_response=wait_for_response)

    async def continueWithAuth(self, requestId: str, authChallengeResponse: list) -> None:
        """
        Продолжает запрос, предоставляющий authChallengeResponse после события authRequired.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-continueWithAuth
        :param requestId:               Идентификатор, полученный клиентом в событии requestPaused.
        :param authChallengeResponse:   Ответ с помощью authChallenge.
                                            {
                                                "response": str(), -> Решение о том, что делать в ответ
                                                    на запрос авторизации. По умолчанию означает
                                                    использование стандартного поведения сетевого стека,
                                                    что, скорее всего, приведет к отмене проверки
                                                    подлинности или появлению всплывающего диалогового
                                                    окна. Возможные значения:
                                                        Default, CancelAuth, ProvideCredentials
                                                "username": str(), -> (optional) Имя пользователя для
                                                    предоставления, может быть пустым. Устанавливается,
                                                    только если ответом является ProvideCredentials.
                                                "password": str(), -> (optional) Пароль пользователя для
                                                    предоставления, может быть пустым. Устанавливается,
                                                    только если ответом является ProvideCredentials.
                                            }
        :return:
        """
        args = {"requestId": requestId, "authChallengeResponse": authChallengeResponse}
        await self._connection.call("Fetch.continueWithAuth", args)

    async def getResponseBody(self, requestId: str) -> dict:
        """
        Вызывает тело ответа, получаемого от сервера и возвращаемого в виде одной строки. Может
            выдаваться только для запроса, который приостановлен на этапе ответа и является
            взаимоисключающим с "takeResponseBodyForInterceptionAsStream". Вызов других методов,
            влияющих на запрос, или отключение домена выборки до получения тела приводит к
            неопределенному поведению.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-getResponseBody
        :param requestId:               Идентификатор перехваченного запроса для получения его тела.
        :return:                        {
                                            "body": str(),        -> Тело ответа.
                                            "base64Encoded": bool -> True - если контент кодирован
                                                                        как base64.
                                        }
        """
        return await self._connection.call("Fetch.getResponseBody", {"requestId": requestId})

    async def takeResponseBodyAsStream(self, requestId: str) -> dict:
        """
        Возвращает дескриптор потока, представляющего тело ответа. Запрос должен быть приостановлен
            на этапе HeadersReceived. Обратите внимание, что после этой команды запрос не может быть
            продолжен как есть - клиент должен либо отменить его, либо предоставить тело ответа.
            Поток поддерживает только последовательное чтение, IO.read потерпит неудачу, если указана
            позиция. Этот метод является взаимоисключающим с getResponseBody. Вызов других методов,
            влияющих на запрос, или отключение домена выборки до получения тела приводит к
            неопределенному поведению.
        https://chromedevtools.github.io/devtools-protocol/tot/Fetch#method-takeResponseBodyAsStream
        :param requestId:               Идентификатор перехваченного запроса для получения его тела.
        :return:                        {
                                            "stream": str(), -> Это либо получается из другого метода,
                                                либо указывается как blob <uuid> это UUID Blob.
                                                    IO.StreamHandle:
                                                    https://chromedevtools.github.io/devtools-protocol/tot/IO#type-StreamHandle
                                        }
        """
        return await self._connection.call("Fetch.takeResponseBodyAsStream", {"requestId": requestId})


class FetchEvent(DomainEvent):
    authRequired = "Fetch.authRequired"
    requestPaused = "Fetch.requestPaused"



try:
    import ujson as json
except ModuleNotFoundError:
    import json

from typing import Optional, List, Dict, Union
from .types import (
    PropertyDescriptor,
    ContextManager,
    ContextDescription,
    Script,
    SerializationOptions,
    RemoteObject,
)
from ...data import DomainEvent
from ...exceptions import (
    PromiseEvaluateError,
    highlight_promise_error,
    EvaluateError,
    highlight_eval_error
)


class Runtime:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Runtime
    """
    __slots__ = ("_connection", "enabled", "context_manager")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled = False
        self.context_manager = ContextManager()

    async def getProperties(
            self, objectId: str,
            skip_complex_types: bool = True,
            ownProperties: Optional[bool] = None,
            accessorPropertiesOnly: Optional[bool] = None,
            generatePreview: Optional[bool] = None,
            nonIndexedPropertiesOnly: Optional[bool] = None,
    ) -> List[PropertyDescriptor]:
        """ Возвращает свойства заданного объекта. Группа объектов результата наследуется
        от целевого объекта.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-getProperties
        :param objectId:                    Идентификатор объекта, для которого возвращаются свойства.
        :param ownProperties:               Если true, возвращает свойства, принадлежащие только самому
                                                элементу, а не его цепочке прототипов.
        :param accessorPropertiesOnly:      Если true, возвращает только свойства доступа (с геттером/сеттером);
                                                внутренние свойства также не возвращаются.
        :param generatePreview:             Должен ли быть создан предварительный просмотр для результатов.
        :param nonIndexedPropertiesOnly:    Если true, возвращает только неиндексированные свойства.
        :return:    {
            "result": array[ PropertyDescriptor ],
            "internalProperties":  list[ InternalPropertyDescriptor ],
            "privateProperties":  list[ PrivatePropertyDescriptor ],
            "exceptionDetails": dict{ ExceptionDetails }
        }
        """
        args = {"objectId": objectId}
        if ownProperties: args.update({"ownProperties": ownProperties})
        if accessorPropertiesOnly: args.update({"accessorPropertiesOnly": accessorPropertiesOnly})
        if generatePreview: args.update({"generatePreview": generatePreview})
        if nonIndexedPropertiesOnly: args.update({"nonIndexedPropertiesOnly": nonIndexedPropertiesOnly})
        response = await self._connection.call("Runtime.getProperties", args)
        if "exceptionDetails" in response:
            raise PromiseEvaluateError(
                highlight_promise_error(response["result"]["description"]) +
                "\n" + json.dumps(response["exceptionDetails"])
            )
        if not skip_complex_types:
            return [PropertyDescriptor(**p) for p in response["result"]]

        result = []
        for p in response["result"]:
            if (subtype := p["value"].get("type")) and subtype == "function":
                continue
            result.append(PropertyDescriptor(**p))
        return result

    async def evaluate(
            self, expression: str,
            objectGroup: Optional[str] = None,
            includeCommandLineAPI: Optional[bool] = None,
            silent: Optional[bool] = None,
            contextId: Optional[int] = None,
            returnByValue: Optional[bool] = None,
            generatePreview: Optional[bool] = None,
            userGesture: Optional[bool] = None,
            awaitPromise: Optional[bool] = None,
            throwOnSideEffect: Optional[bool] = None,
            timeout: Optional[float] = None,
            disableBreaks: Optional[bool] = None,
            replMode: Optional[bool] = None,
            allowUnsafeEvalBlockedByCSP: Optional[bool] = None,
            uniqueContextId: Optional[str] = None,
            serializationOptions: Optional[SerializationOptions] = None,
    ) -> RemoteObject:
        """ Выполняет JavaScript-выражение в глобальном контексте.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime/#method-evaluate
        :param expression:  JavaScript-выражение.
        :param objectGroup: Символическое имя группы, которое можно использовать
            для освобождения нескольких объектов.
        :param includeCommandLineAPI:   Определяет, должен ли быть доступен API
            командной строки во время выполнения выражения.
        :param silent:  В автоматическом режиме исключения, возникающие во время
            выполнения выражения, не сообщаются и не приостанавливают выполнение.
            Переопределяет setPauseOnException состояние.
        :param contextId:   Указывает, в каком контексте выполнения выполнять выражение.
            Если параметр опущен, оценка будет выполняться в контексте проверяемой
            страницы. Это взаимоисключающее свойство uniqueContextId, которое предлагает
            альтернативный способ определения контекста выполнения, более надежный в
            многопроцессорной среде.
        :param returnByValue:   Ожидается ли, что результат будет объектом JSON, который
            должен быть отправлен по значению.
        :param generatePreview: Должен ли быть создан предварительный просмотр для результата.
        :param userGesture: Следует ли рассматривать выполнение как инициированное
            пользователем в пользовательском интерфейсе.
        :param awaitPromise:    Должно ли выполнение выражения ожидать результирующего
            значения и возвращаться после завершения промиса.
        :param throwOnSideEffect:   Выбрасывать ли исключение, если во время выполнения
            выражения нельзя исключить побочный эффект. Это подразумевает использование
             disableBreaks аргумента.
        :param timeout: Прервать выполнение по истечении времени ожидания (количество миллисекунд).
        :param disableBreaks:   Отключите точки останова во время выполнения выражения.
        :param replMode:    Установка этого флага в значение true разрешает повторное
            объявление переменных используя 'let', а так же вызов awaitable инструкций
            в top-level контекста. Обратите внимание, что переменные let могут быть
            повторно объявлены только в том случае, если они сами происходят из replMode.
        :param allowUnsafeEvalBlockedByCSP: Политика безопасности содержимого (CSP)
            для target может блокировать «unsafe-eval», который включает eval(),
            Function(), setTimeout() и setInterval() при вызове с аргументами, которые
            нельзя вызывать. Этот флаг обходит CSP для этой оценки и разрешает
            небезопасную оценку. По умолчанию true.
        :param uniqueContextId: Альтернативный способ указать контекст выполнения
            для выполнения выражения. По сравнению с contextId, который может повторно
            использоваться в процессах, он гарантированно уникален для системы, поэтому
            его можно использовать для предотвращения случайного вычисления выражения
            в контексте, отличном от предполагаемого (например, в результате навигации
            через границы процесса). Это взаимоисключающее значение с contextId.
        :param serializationOptions:    Указывает сериализацию результата. Если указано,
            переопределяет generatePreview, returnByValue.
        :return:
        """
        args = {"expression": expression}
        if objectGroup is not None: args.update(objectGroup=objectGroup)
        if includeCommandLineAPI is not None:
            args.update(includeCommandLineAPI=includeCommandLineAPI)
        if silent is not None: args.update(silent=silent)
        if contextId is not None: args.update(contextId=contextId)
        if returnByValue is not None: args.update(returnByValue=returnByValue)
        if generatePreview is not None: args.update(generatePreview=generatePreview)
        if userGesture is not None: args.update(userGesture=userGesture)
        if awaitPromise is not None: args.update(awaitPromise=awaitPromise)
        if throwOnSideEffect is not None: args.update(throwOnSideEffect=throwOnSideEffect)
        if timeout is not None: args.update(timeout=timeout)
        if disableBreaks is not None: args.update(disableBreaks=disableBreaks)
        if replMode is not None: args.update(replMode=replMode)
        if allowUnsafeEvalBlockedByCSP is not None:
            args.update(allowUnsafeEvalBlockedByCSP=allowUnsafeEvalBlockedByCSP)
        if uniqueContextId is not None: args.update(uniqueContextId=uniqueContextId)
        if serializationOptions is not None:
            args.update(serializationOptions=serializationOptions.as_dict())

        response = await self._connection.call("Runtime.evaluate", args)
        if "exceptionDetails" in response:
            raise EvaluateError(
                highlight_eval_error(response["result"]["description"], expression)
            )

        return RemoteObject(**response["result"])

    async def awaitPromise(
            self, promiseObjectId: str, returnByValue: bool = False, generatePreview: bool = False
    ) -> RemoteObject:
        """ Дожидается выполнения промиса с указанным promiseObjectId.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-awaitPromise
        :param promiseObjectId:     Идентификатор промиса.
        :param returnByValue:       (optional) Ожидается ли результат в виде объекта JSON,
                                        который должен быть отправлен по значению.
        :param generatePreview:     (optional) Должен ли предварительный просмотр
                                        генерироваться для результата.
        :return:
        """
        args = {"promiseObjectId": promiseObjectId, "returnByValue": returnByValue, "generatePreview": generatePreview}
        response = await self._connection.call("Runtime.awaitPromise", args)
        if "exceptionDetails" in response:
            raise PromiseEvaluateError(
                highlight_promise_error(response["result"]["description"]) +
                "\n" + json.dumps(response["exceptionDetails"])
            )
        return RemoteObject(**response["result"])

    async def callFunctionOn(
            self, functionDeclaration: str,
            objectId: Optional[str] = None,
            arguments: Optional[list] = None,
            silent: Optional[bool] = None,
            returnByValue: Optional[bool] = None,
            generatePreview: Optional[bool] = None,
            userGesture: Optional[bool] = None,
            awaitPromise: Optional[bool] = None,
            executionContextId: Optional[int] = None,
            objectGroup: Optional[str] = None,
            throwOnSideEffect: Optional[bool] = None,
            uniqueContextId: Optional[str] = None,
            serializationOptions: Optional[SerializationOptions] = None
    ) -> RemoteObject:
        """
        Вызывает функцию с заданным объявлением для данного объекта. Группа объектов результата
            наследуется от целевого объекта.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-callFunctionOn
        :param functionDeclaration:     Объявление функции для вызова.
        :param objectId:                (optional) Идентификатор объекта для вызова функции.
                                            Должен быть указан либо objectId, либо executeContextId.
        :param arguments:               (optional) Аргументы. Все аргументы вызова должны
                                            принадлежать тому же миру JavaScript, что и целевой
                                            объект.
        :param silent:                  (optional) В тихом режиме исключения, выданные во время оценки,
                                            не регистрируются и не приостанавливают выполнение.
                                            Переопределяет 'setPauseOnException' состояние.
        :param returnByValue:           (optional) Ожидается ли результат в виде объекта JSON,
                                            который должен быть отправлен по значению.
        :param generatePreview:         (optional, EXPERIMENTAL) Должен ли предварительный
                                            просмотр генерироваться для результата.
        :param userGesture:             (optional) Должно ли выполнение рассматриваться как
                                            инициированное пользователем в пользовательском интерфейсе.
        :param awaitPromise:            (optional) Решено ли выполнение await для полученного значения
                                            и возврата после ожидаемого обещания.
        :param executionContextId:      (optional) Определяет контекст выполнения, в котором будет
                                            использоваться глобальный объект для вызова функции.
                                            Должен быть указан либо executeContextId, либо objectId.
        :param objectGroup:             (optional) Символическое имя группы, которое можно
                                            использовать для освобождения нескольких объектов. Если
                                            objectGroup не указан, а objectId равен, objectGroup
                                            будет унаследован от объекта.
        :param throwOnSideEffect:   Выбрасывать ли исключение, если во время выполнения функции
            нельзя исключить побочный эффект.
        :param uniqueContextId: Альтернативный способ указать контекст выполнения
            для выполнения выражения. По сравнению с contextId, который может повторно
            использоваться в процессах, он гарантированно уникален для системы, поэтому
            его можно использовать для предотвращения случайного вычисления выражения
            в контексте, отличном от предполагаемого (например, в результате навигации
            через границы процесса). Это взаимоисключающее значение с contextId.
        :param serializationOptions:    Указывает сериализацию результата. Если указано,
            переопределяет generatePreview, returnByValue.
        :return:
        """
        args = {"functionDeclaration": functionDeclaration}
        if objectId is not None:
            args.update({"objectId": objectId})
        if arguments is not None:
            args.update({"arguments": arguments})
        if silent is not None:
            args.update({"silent": silent})
        if returnByValue is not None:
            args.update({"returnByValue": returnByValue})
        if generatePreview is not None:
            args.update({"generatePreview": generatePreview})
        if userGesture is not None:
            args.update({"userGesture": userGesture})
        if awaitPromise is not None:
            args.update({"awaitPromise": awaitPromise})
        if executionContextId is not None:
            args.update({"executionContextId": executionContextId})
        if objectGroup is not None:
            args.update({"objectGroup": objectGroup})
        if throwOnSideEffect is not None:
            args.update(throwOnSideEffect=throwOnSideEffect)
        if uniqueContextId is not None:
            args.update(uniqueContextId=uniqueContextId)
        if serializationOptions is not None:
            args.update(serializationOptions=serializationOptions.as_dict())
        response = await self._connection.call("Runtime.callFunctionOn", args)
        if "exceptionDetails" in response:
            raise EvaluateError(
                highlight_eval_error(response["result"]["description"], functionDeclaration)
            )
        return RemoteObject(**response["result"])

    async def enable(self, watch_for_execution_contexts: bool = False) -> None:
        """
        Включает создание отчетов о создании контекстов выполнения с помощью события executeContextCreated.
            При включении, событие будет отправлено немедленно для каждого существующего контекста выполнения.

        Позволяет так же организовать обратную связь со страницей, посылая из её контекста, данные, в консоль.
            В этом случае будет генерироваться событие 'Runtime.consoleAPICalled':
            https://chromedevtools.github.io/devtools-protocol/tot/Runtime#event-consoleAPICalled
            {
                'method': 'Runtime.consoleAPICalled',
                'params': {
                    'type': 'log',
                    'args': [{'type': 'string', 'value': 'you console data passed was be here'}],
                    'executionContextId': 2,
                    'timestamp': 1583582949679.153,
                    'stackTrace': {
                        'callFrames': [{'functionName': '', 'scriptId': '48', 'url': '', 'lineNumber': 0, 'columnNumber': 8}]
                    }
                }
            }

        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-enable
        :param watch_for_execution_contexts:    Регистрирует слушателей, ожидающих события создания/уничтожения
                                                    контекстов, которые можно запрашивать через
                                                    page_instance.context_manager.GetDefaultContext(frameId: str).
                                                    Должен быть включён ПЕРЕД переходом на целевой адрес.
        :return:
        """
        if not self.enabled:
            await self._connection.call("Runtime.enable")
            self.enabled = True

        if watch_for_execution_contexts and not self.context_manager.is_watch:
            await self._connection.addListenerForEvent(
                RuntimeEvent.executionContextCreated, self.context_manager.on_create)
            await self._connection.addListenerForEvent(
                RuntimeEvent.executionContextsCleared, self.context_manager.on_clear)
            await self._connection.addListenerForEvent(
                RuntimeEvent.executionContextDestroyed, self.context_manager.on_destroy)
            self.context_manager.is_watch = True

    async def disable(self) -> None:
        """
        Отключает создание отчетов о создании контекстов выполнения.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-disable
        :return:
        """
        if self.enabled:
            await self._connection.call("Runtime.disable")
            self.enabled = False

        if self.context_manager.is_watch:
            self._connection.removeListenerForEvent(
                RuntimeEvent.executionContextCreated, self.context_manager.on_create)
            self._connection.removeListenerForEvent(
                RuntimeEvent.executionContextsCleared, self.context_manager.on_clear)
            self._connection.removeListenerForEvent(
                RuntimeEvent.executionContextDestroyed, self.context_manager.on_destroy)
            self.context_manager.is_watch = False

    async def discardConsoleEntries(self) -> None:
        """ Отбрасывает собранные исключения и вызовы API консоли.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-discardConsoleEntries
        :return:
        """
        await self._connection.call("Runtime.discardConsoleEntries")

    async def releaseObjectGroup(self, objectGroup: str) -> None:
        """ Освобождает все удаленные объекты, принадлежащие данной группе.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-releaseObjectGroup
        :param objectGroup:             Символическое имя группы.
        :return:
        """
        await self._connection.call("Runtime.releaseObjectGroup", dict(objectGroup=objectGroup))

    async def releaseObject(self, objectId: str) -> None:
        """  Освобождает удаленный объект, с указанным objectId.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-releaseObject
        :param objectId:             Символическое имя группы.
        :return:
        """
        await self._connection.call("Runtime.releaseObject", dict(objectId=objectId))

    async def compileScript(
            self, expression: str,
            sourceURL: str = "",
            persistScript: bool = True,
            executionContextId: Optional[int] = None
    ) -> str:
        """ Компилирует выражение.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-compileScript
        :param expression:              Выражение для компиляции.
        :param sourceURL:               Исходный URL для скрипта.
        :param persistScript:           Указывает, следует ли сохранить скомпилированный скрипт.
        :param executionContextId:      (optional) Указывает, в каком контексте выполнения выполнять сценарий.
                                            Если параметр не указан, выражение будет выполняться в контексте
                                            проверяемой страницы.
        :return:
        """
        args: Dict[str, Union[int, str, bool]] = dict(
            expression=expression, sourceURL=sourceURL, persistScript=persistScript)
        if executionContextId is not None:
            args.update(executionContextId=executionContextId)

        response = await self._connection.call("Runtime.compileScript", args)
        if "exceptionDetails" in response:
            raise EvaluateError(
                highlight_eval_error(response["result"]["description"], expression)
            )
        return response["scriptId"]

    async def buildScript(self, expression: str, context: Optional[ContextDescription] = None) -> Script:
        return Script(self._connection, expression, context)

    async def runIfWaitingForDebugger(self) -> None:
        """ Сообщает инспектируемой странице, что можно запуститься, если она ожидает этого после
        Target.setAutoAttach.
        """
        await self._connection.call("Runtime.runIfWaitingForDebugger")

    async def runScript(
            self, scriptId: str,
            executionContextId: Optional[int] = None,
            objectGroup: str = "console",
            silent: bool = False,
            includeCommandLineAPI: bool = True,
            returnByValue: bool = False,
            generatePreview: bool = False,
            awaitPromise: bool = True
    ) -> RemoteObject:
        """
        Запускает скрипт с заданным идентификатором в заданном контексте.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-runScript
        :param scriptId:                ID сценария для запуска.
        :param executionContextId:      (optional) Указывает, в каком контексте выполнения выполнять сценарий.
                                            Если параметр не указан, выражение будет выполняться в контексте
                                            проверяемой страницы.
        :param objectGroup:             (optional) Символическое имя группы, которое можно использовать для
                                            освобождения нескольких объектов.
        :param silent:                  (optional) В тихом режиме исключения, выданные во время оценки, не
                                            сообщаются и не приостанавливают выполнение. Переопределяет
                                            состояние setPauseOnException.
        :param includeCommandLineAPI:   (optional) Определяет, должен ли API командной строки быть доступным
                                            во время оценки.
        :param returnByValue:           (optional) Ожидается ли результат в виде объекта JSON, который должен
                                            быть отправлен по значению.
        :param generatePreview:         (optional) Должен ли предварительный просмотр генерироваться для результата.
        :param awaitPromise:            (optional) Будет ли выполнено ожидание выполнения для полученного значения
                                            и возврата после ожидаемого 'promise'.
        :return:                        {
                                            "result": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-RemoteObject)
                                            "exceptionDetails": dict(https://chromedevtools.github.io/devtools-protocol/tot/Runtime#type-ExceptionDetails)
                                        }
        """
        args = {
            "scriptId": scriptId, "objectGroup": objectGroup, "silent": silent,
            "includeCommandLineAPI": includeCommandLineAPI, "returnByValue": returnByValue,
            "generatePreview": generatePreview, "awaitPromise": awaitPromise
        }
        if executionContextId is not None:
            args.update({"executionContextId": executionContextId})

        response = await self._connection.call("Runtime.runScript", args)
        if "exceptionDetails" in response:
            raise EvaluateError(
                highlight_eval_error(response["result"]["description"], "scriptId: " + scriptId)
            )
        return RemoteObject(**response["result"])

    async def addBinding(self, name: str, executionContextName: Optional[str] = None) -> None:
        """ Делает доступным переданное имя в качестве имени функции, доступной глобально. Вызов
        такой функции принимает ровно один аргумент типа `string`, иначе возбуждается исключение
        с текстом `Invalid arguments: should be exactly one string.`. Каждый вызов функции по
        этому имени создает уведомление Runtime.bindingCalled, в теле которого, поле `payload`
        будет содержать строку, переданную в качестве аргумента вызванной функции.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-addBinding
        :param name:                    Имя вызываемой функции.
        :param executionContextName:    (optional) Имя контекста исполнения.
        :return:
        """
        args = {"name": name}
        if executionContextName is not None:
            args.update({"executionContextName": executionContextName})

        await self._connection.call("Runtime.addBinding", args)

    async def removeBinding(self, name: str) -> None:
        """ Удаляет привязку по имени. Функцию всё ещё можно будет вызвать, но событие
        'Runtime.bindingCalled' возбуждаться  не будет.
        https://chromedevtools.github.io/devtools-protocol/tot/Runtime#method-removeBinding
        :param name:                    Имя вызываемой функции.
        :return:
        """
        await self._connection.call("Runtime.removeBinding", dict(name=name))


class RuntimeEvent(DomainEvent):
    consoleAPICalled = "Runtime.consoleAPICalled"
    exceptionRevoked = "Runtime.exceptionRevoked"
    exceptionThrown = "Runtime.exceptionThrown"
    executionContextCreated = "Runtime.executionContextCreated"
    executionContextDestroyed = "Runtime.executionContextDestroyed"
    executionContextsCleared = "Runtime.executionContextsCleared"
    inspectRequested = "Runtime.inspectRequested"
    bindingCalled = "Runtime.bindingCalled"                       # ! EXPERIMENTAL



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class ProcessInfo:
    type: str  # Тип процесса
    id: int  # Идентификатор процесса
    cpuTime: float  # Совокупное использование ЦП в секундах для всех потоков процесса с момента его запуска.


@dataclass
class GPUInfo:
    devices: List[dict]
    driverBugWorkarounds: List[str]
    videoDecoding: List[dict]
    videoEncoding: List[dict]
    imageDecoding: List[dict]
    auxAttributes: Optional[dict] = None
    featureStatus: Optional[dict] = None


@dataclass
class SystemData:
    gpu: "GPUInfo"
    modelName: str
    modelVersion: str
    commandLine: str




__version__ = "1.1.1"
__author__ = "PieceOfGood"
__email__ = "78sanchezz@gmail.com"

__all__ = [
    "find_instances",
    "CMDFlags",
    "FlagBuilder",
    "Browser",
    "Connection",
    "BrowserName"
]

from .browser import CMDFlags
from .browser import FlagBuilder
from .browser import Browser
from .connection import Connection
from .utils import find_instances


class BrowserName:
    CHROME = "chrome"
    CHROMIUM = "chromium"
    BRAVE = "brave"
    EDGE = "edge"



import asyncio
import random
from typing import Tuple, Optional, Literal
from .data import WINDOWS_KEY_SET, KeyModifiers, KeyEvents
from .domains.browser.types import Bounds


class Actions:
    __slots__ = ("_connection",)

    def __init__(self, conn) -> None:
        from .connection import Connection

        self._connection: Connection = conn

    async def swipeTo(
        self,
        direction: Literal["up", "down", "left", "right"],
        x:            float = 0,
        y:            float = 0,
        distance:     Optional[float] = None,
        speed:        Optional[float] = None,
        overscroll:   Optional[float] = None,
        repeat_count:   Optional[int] = None,
        repeat_delay: Optional[float] = None,
        gestureSourceType: Literal["touch", "mouse"] = "mouse"
    ) -> None:
        """
        Скроллит вьюпорт жестом "touch" на всю его длину/высоту.
            Возвращает управление только после выполнения жеста!
        :param direction:           (optional) Направление. Может быть следующим:
                                        up — пальцем вверх(прокрутка вниз)
                                        down — пальцем вниз(прокрутка вверх)
                                        left — пальцем влево(прокрутка вправо)
                                        right — пальцем вправо(прокрутка влево)
        :param x:                   (optional) X-координата начальной точки.
        :param y:                   (optional) Y-координата начальной точки.
        :param distance:            (optional) Дистанция движения.
        :param speed:               (optional) Скорость движения(по умолчанию = 800).
        :param overscroll:          (optional) Дополнительная дистанция в пикселях.
        :param repeat_count:        (optional) Кол-во повторений сделанного жеста.
        :param repeat_delay:        (optional) Задержка между повторениями.
        :param gestureSourceType:   (optional) Задержка между повторениями.
        :return:
        """
        if direction not in ["up", "down", "left", "right"]:
            raise ValueError("'direction' must be one from — up; down; left; right")
        sign = -1 if direction in ["up", "left"] else 1
        rect = None
        if x is None:
            rect = await self._connection.extend.getViewportRect()
            x = 10 if direction == "right" else rect.width - 10 if direction == "left" else rect.height / 2

        if y is None:
            rect = await self._connection.extend.getViewportRect()
            y = 10 if direction == "down" else rect.height - 10 if direction == "up" else rect.width / 2

        if distance is None:
            rect = rect or await self._connection.extend.getViewportRect()
            distance = (rect.height if direction in ["up", "down"] else rect.width) - 10

        overscroll = overscroll if overscroll is not None else 0

        args = {
            "x": x, "y": y, "speed": speed, "repeatCount": repeat_count,
            "repeatDelayMs": repeat_delay, "gestureSourceType": gestureSourceType,
            "xDistance": distance * sign if direction in ["left", "right"] else None,
            "yDistance": distance * sign if direction in ["up", "down"] else None,
            "xOverscroll": overscroll * -sign if direction in ["left", "right"] else None,
            "yOverscroll": overscroll * -sign if direction in ["up", "down"] else None,
        }
        await self._connection.Input.synthesizeScrollGesture(**args)

    async def clickTo(self, x: int, y: int, delay: float = None) -> None:
        """
        Эмулирует клик мыши по координатам.
        :param x:               x - координата
        :param y:               y - координата
        :param delay:           задержка перед отпусканием
        :return:
        """
        await self._connection.Input.dispatchMouseEvent("mousePressed", x, y, button="left")
        if delay: await asyncio.sleep(delay)
        await self._connection.Input.dispatchMouseEvent("mouseReleased", x, y, button="left")

    async def mouseMoveTo(self, x: int, y: int) -> None:
        await self._connection.Input.dispatchMouseEvent("mouseMoved", x, y)

    async def mouseWheel(self, x: float, y: float, deltaX: float = 0, deltaY: float = 0) -> None:
        """
        Крутит колесо мышки.
        :param x:               Положение
        :param y:               указателя мыши.
        :param deltaX:          Скроллит по-горизонтали.
        :param deltaY:          Скроллит по-вертикали, положительное значение == вниз.
        :return:
        """
        await self._connection.Input.dispatchMouseEvent("mouseWheel", x, y, button="middle", deltaX=deltaX, deltaY=deltaY)

    async def wheelTo(self, direction: str = "down") -> None:
        sign = -1 if direction in ["up", "left"] else 1
        rect = await self._connection.extend.getViewportRect()
        x = 10 if direction == "right" else rect.width - 10 if direction == "left" else rect.height / 2
        y = 10 if direction == "down" else rect.height - 10 if direction == "up" else rect.width / 2
        distance = ((rect.height if direction in ["up", "down"] else rect.width) - 10) * sign
        delta = {"deltaX": distance} if direction in ["left", "rigth"] else {"deltaY": distance}
        await self.mouseWheel(x, y, **delta)


    async def mouseMoveToCoordinatesAndClick(self, x: int, y: int) -> None:
        await self.mouseMoveTo(x, y)
        await self.clickTo(x, y)

    async def sendChar(self, char: str) -> None:
        """
        Эмулирует ввод символа нажатием соответствующей кнопки клавиатуры.
            Внимание! Курсор должен быть установлен в эдит-боксе!
        :param char:             Символ для ввода.
        :return:
        """
        upper_key = char.upper()
        args = {
            "text": char, "key": char, "keyIdentifier": f"U+{WINDOWS_KEY_SET[upper_key]:X}",
            "windowsVirtualKeyCode": WINDOWS_KEY_SET[upper_key],
            "nativeVirtualKeyCode": WINDOWS_KEY_SET[upper_key]
        }
        if len(char) > 1: raise ValueError(f"Передаваемая строка: '{char}' — должна быть из одного символа!")
        await self._connection.Input.dispatchKeyEvent("char", **args)

    async def sendText(
            self, text: str, interval: Optional[Tuple[float, float]] = None
    ) -> None:
        """
        Эмулирует последовательный набор текста.
            Внимание! Курсор должен быть установлен в эдит-боксе!
        :param text:             Последовательность символов для ввода.
        :param interval:         Задержка после нажатия кнопки. None - отключает задержку.
                                     Кортеж из двух чисел описывает интервал, вычисляемый рандомно.
                                     Кортеж (10, None) — устанавливает фиксированное ожидание в 10 секунд
        :return:
        """
        for letter in text:
            await self.sendChar(letter)
            if interval is not None: await asyncio.sleep(random.uniform(interval[0], interval[1]))

    async def sendKeyEvent(self, event: dict, *modifiers: KeyModifiers) -> None:
        args = {}
        if modifiers:
            s = 0
            for m in modifiers: s += m.value
            args.update({"modifiers": s})

        args.update(event)
        await self._connection.Input.dispatchKeyEvent("keyDown", **args)
        await self._connection.Input.dispatchKeyEvent("keyUp", **args)

    async def controlA(self) -> None:
        """ Выделить весь текст(Ctrl+A). """
        await self.sendKeyEvent(KeyEvents.keyA, KeyModifiers.ctrl)

    async def backspaceText(
            self, count: int = 1, modifier: KeyModifiers = KeyModifiers.none) -> None:
        """
        Удаляет текст в текстовом поле с позиции курсора по направлению в лево, или полностью очистить,
            эмулируя нажатие клавиши 'Backspace'.
            Внимание! Курсор должен быть установлен в эдит-боксе!
        :param count:                   (optional) Количество нажатий. Не имеет воздействия при полной очистке.
        :param modifier:                (optional) none - удалить один символ, alt - слово, включая стоящие перед
                                            ним пробелы, ctrl — полностью очистить эдит-бокс.
        :return:
        """

        if modifier.name == "ctrl":
            await self.controlA()
            await self.sendKeyEvent(KeyEvents.backspace)
            return

        for i in range(count):
            await self.sendKeyEvent(KeyEvents.backspace, modifier.value)

    async def setWindowBounds(self, bounds: Bounds, windowId: Optional[int] = None) -> None:
        """
        (EXPERIMENTAL)
        Устанавливает позицию и/или размер окна.
        https://chromedevtools.github.io/devtools-protocol/tot/Browser#method-setWindowBounds
        :param bounds:          Новые границы окна, а так же состояние.
        :param windowId:        Идентификатор окна.

        :return:        None
        """
        if windowId is None:
            windowId = (await self._connection.Target.getWindowForTarget()).windowId
        await self._connection.call("Browser.setWindowBounds", {"windowId": windowId, "bounds": bounds.to_dict()})



from dataclasses import dataclass, field
from typing import Optional, List, Literal, Union


@dataclass
class LifecycleEventData:
    frameId: str
    loaderId: str
    name: str
    timestamp: float


@dataclass
class AdFrameStatus:
    """ Указывает, был ли frame идентифицирован как реклама и почему. """
    adFrameType: Literal["none", "child", "root"]
    explanations: Optional[List[Literal["ParentIsAd", "CreatedByAdScript", "MatchedBlockingRule"]]] = None


@dataclass
class Frame:
    """ Информация о фрейме на странице """
    id: str
    loaderId: str                               # ? Network.LoaderId
    url: str
    domainAndRegistry: str
    securityOrigin: str
    mimeType: str
    secureContextType: Literal["Secure", "SecureLocalhost", "InsecureScheme", "InsecureAncestor"]
    crossOriginIsolatedContextType: Literal["Isolated", "NotIsolated", "NotIsolatedFeatureDisabled"]
    gatedAPIFeatures: List[Literal[
        "SharedArrayBuffers", "SharedArrayBuffersTransferAllowed", "PerformanceMeasureMemory", "PerformanceProfile"
    ]]
    adFrameStatus: Optional['AdFrameStatus']
    parentId: Optional[str] = None
    name: Optional[str] = None
    urlFragment: Optional[str] = None
    unreachableUrl: Optional[str] = None
    _adFrameStatus: Optional['AdFrameStatus'] = field(init=False, repr=False, default=None)

    @property
    def adFrameStatus(self) -> Union['AdFrameStatus', None]:
        return self._adFrameStatus

    @adFrameStatus.setter
    def adFrameStatus(self, data: dict) -> None:
        self._adFrameStatus = AdFrameStatus(**data) if not isinstance(data, property) else None


@dataclass
class FrameTree:
    """ Информация об иерархии фреймов """
    frame: 'Frame'
    childFrames: Optional[List['FrameTree']]
    _childFrames: Optional[List['FrameTree']] = field(init=False, repr=False, default=None)

    @property
    def frame(self) -> 'Frame':
        return self._frame

    @frame.setter
    def frame(self, data: dict) -> None:
        self._frame = Frame(**data)

    @property
    def childFrames(self) -> Union[List['FrameTree'], None]:
        return self._childFrames

    @childFrames.setter
    def childFrames(self, data: List[dict]) -> None:
        if not isinstance(data, property):
            self._childFrames = [FrameTree(**frame_data) for frame_data in data]
        else:
            self._childFrames = None

    def GetFrameByUrl(self, value: str) -> Union['Frame', None]:
        def search(tree: 'FrameTree', value: str) -> Union['Frame', None]:
            # print("tree.frame.url", tree.frame.url)
            if value in tree.frame.url:
                return tree.frame
            else:
                if tree.childFrames:
                    for child in tree.childFrames:
                        if result := search(child, value):
                            return result
            return None
        return search(self, value)



import re
from typing import Optional, Union, List
from .dom_element import Node
from ..runtime.types import RemoteObject
from ...exceptions import CouldNotFindNodeWithGivenID, RootIDNoLongerExists
from ...data import DomainEvent


class DOM:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/DOM
    """
    __slots__ = ("_connection", "enabled")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.enabled = False

    async def enable(self) -> None:
        """
        Включает DOM-агент для данной страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-enable
        :return:
        """
        if not self.enabled:
            await self._connection.call("DOM.enable")
            self.enabled = True

    async def disable(self) -> None:
        """
        Отключает DOM-агент для данной страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-disable
        :return:
        """
        if self.enabled:
            await self._connection.call("DOM.disable")
            self.enabled = False

    async def getRoot(self, depth: Optional[int] = None, pierce: Optional[bool] = None) -> Node:
        """
        Возвращает корневой узел документа.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getDocument
        Корневой элемент ВСЕГДА имеет подобную структуру:
        'root': {
            'nodeId': 19,
            'backendNodeId': 2,
            'nodeType': 9,
            'nodeName': '#document',
            'localName': '',
            'nodeValue': '',
            'childNodeCount': 2,
            'children': [
                {
                    'nodeId': 20,
                    'parentId': 19,
                    'backendNodeId': 9,
                    'nodeType': 10,
                    'nodeName': 'html',
                    'localName': '',
                    'nodeValue': '',
                    'publicId': '',
                    'systemId': ''
                }, {
                    'nodeId': 21,
                    'parentId': 19,
                    'backendNodeId': 10,
                    'nodeType': 1,
                    'nodeName': 'HTML',
                    'localName': 'html',
                    'nodeValue': '',
                    'childNodeCount': 2,
                    'children': [
                        {
                            'nodeId': 22,
                            'parentId': 21,
                            'backendNodeId': 11,
                            'nodeType': 1,
                            'nodeName': 'HEAD',
                            'localName': 'head',
                            'nodeValue': '',
                            'childNodeCount': 4,
                            'attributes': [ ]
                        }, {
                            'nodeId': 23,
                            'parentId': 21,
                            'backendNodeId': 12,
                            'nodeType': 1,
                            'nodeName': 'BODY',
                            'localName': 'body',
                            'nodeValue': '',
                            'childNodeCount': 8,
                            'attributes': [ ]
                        }
                    ],
                    'attributes': [
                        'lang',
                        'ru'
                    ],
                    'frameId': 'AF11E1D7BC9DF951D77C6C07C02B98E7'
                }
            ],
            'documentURL': 'url ...',
            'baseURL': 'url ...',
            'xmlVersion': ''
        }
        :param depth:           Максимальная глубина, на которой должны быть извлечены
                                    дочерние элементы, по умолчанию равна 1. Используйте
                                    -1 для всего поддерева или укажите целое число больше 0.
        :param pierce:          Должны ли проходиться iframes и теневые корни при возврате
                                    поддерева (по умолчанию false).
        :return:            <Node>.
        """
        args = {}
        if depth is not None: args.update(depth=depth)
        if pierce is not None: args.update(pierce=pierce)
        node: dict = (await self._connection.call("DOM.getDocument", args))["root"]
        return Node(self._connection, **node)

    async def querySelector(
            self, selector: str,
            ignore_root_id_exists: bool = False,
            in_frames: bool = False
    ) -> Union[Node, None]:
        """
        Выполняет DOM-запрос, возвращая объект найденного узла, или None.
            Эквивалент  === document.querySelector()
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-querySelector
        :param selector:                    Селектор.
        :param ignore_root_id_exists:       Игнорировать исключение при отсутствии родительского элемента.
                                                Полезно при запросах на загружающихся страницах.
        :param in_frames:                   Опрашивать документ вкючая shadow-root и iframe?
        :return:                <Node>
        """
        args = {} if not in_frames else dict(depth=-1, pierce=True)
        root_node_id = (await self._connection.call("DOM.getDocument", args))["root"]["nodeId"]
        try:
            node: dict = await self._connection.call("DOM.querySelector", {
                "nodeId": root_node_id, "selector": selector
            })
        except CouldNotFindNodeWithGivenID as e:
            if match := re.search(r"nodeId\': (\d+)", str(e)):
                if match.group(1) == str(root_node_id):
                    if ignore_root_id_exists:
                        return None
                    raise RootIDNoLongerExists
            raise
        return Node(self._connection, **node) if node["nodeId"] > 0 else None


    async def querySelectorAll(
            self, selector: str,
            ignore_root_id_exists: bool = False,
            in_frames: bool = False
    ) -> List[Node]:
        """
        Выполняет DOM-запрос, возвращая список объектов найденных узлов, или пустой список.
            Эквивалент  === document.querySelectorAll()
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-querySelectorAll
        :param selector:                    Селектор.
        :param ignore_root_id_exists:       Игнорировать исключение при отсутствии родительского элемента.
                                                Полезно при запросах на загружающихся страницах.
        :param in_frames:                   Опрашивать документ вкючая shadow-root и iframe?
        :return:                [ <Node>, <Node>, ... ]
        """
        nodes = []
        args = {} if not in_frames else dict(depth=-1, pierce=True)
        root_node_id = (await self._connection.call("DOM.getDocument", args))["root"]["nodeId"]
        try:
            for node in (await self._connection.call("DOM.querySelectorAll", {
                "nodeId": root_node_id, "selector": selector
            }))["nodeIds"]:
                nodes.append(Node(self._connection, node))
        except CouldNotFindNodeWithGivenID as e:
            if match := re.search(r"nodeId\': (\d+)", str(e)):
                if match.group(1) == str(root_node_id):
                    if ignore_root_id_exists:
                        return []
                    raise RootIDNoLongerExists
            raise
        return nodes

    async def performSearch(self, query: str, searchInShadowDOM: Optional[bool] = None) -> dict:
        """
        (EXPERIMENTAL)
        Ищет заданную строку в дереве DOM. Используйте 'GetSearchResults()' для доступа к результатам
            поиска или 'CancelSearch()'( !не найдено! ), чтобы завершить этот сеанс поиска. DOM-агент
            должен быть влючён.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-performSearch
        :param query:               Обычный текст, селектор, или поисковый запрос XPath.
        :param searchInShadowDOM:   (optional) True - поиск будет так же выполнен в shadow DOM.
        :return:                    {"searchId": str(searchId), "resultCount": int(resultCount)}
                                        searchId    - уникальный идентификатор сессии поиска.
                                        resultCount - кол-во результатов удовлетворяющих запрос.
        """
        args = {"query": query}
        if searchInShadowDOM is not None:
            args.update({"includeUserAgentShadowDOM": searchInShadowDOM})
        return await self._connection.call("DOM.performSearch", args)

    async def getSearchResults(
            self, searchId: str,
            fromIndex: int = 0,
            toIndex:   int = 0
    ) -> List[Node]:
        """
        (EXPERIMENTAL)
        Возвращает список результатов поиска для поисковой сессии 'searchId', в интервале от 'fromIndex'
            до 'toIndex', полученной в результате вызова PerformSearch().
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-getSearchResults
        :param searchId:        Уникальный идентификатор сессии поиска.
        :param fromIndex:       Начальный индекс результата поиска, который будет возвращен.
        :param toIndex:         Конечный индекс результата поиска, который будет возвращен.
        :return:                [ <Node>, <Node>, ... ]
        """
        nodes = []
        args = {"searchId": searchId, "fromIndex": fromIndex, "toIndex": toIndex}
        for node_id in (await self._connection.call("DOM.getSearchResults", args))["nodeIds"]:
            nodes.append(Node(self._connection, node_id))
        return nodes

    async def undo(self) -> None:
        """
        (EXPERIMENTAL)
        Отменяет последнее выполненное действие.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM#method-undo
        :return:
        """
        await self._connection.call("DOM.undo")

    async def redo(self) -> None:
        """
        (EXPERIMENTAL)
        Повторно выполняет последнее отмененное действие.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-redo
        :return:
        """
        await self._connection.call("DOM.redo")

    async def markUndoableState(self) -> None:
        """
        (EXPERIMENTAL)
        Отмечает последнее состояние, которое нельзя изменить.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-markUndoableState
        :return:
        """
        await self._connection.call("DOM.markUndoableState")

    async def describeNode(
            self, nodeId: Optional[int] = None,
            backendNodeId: Optional[int] = None,
            objectId: Optional[str] = None,
            depth: Optional[int] = None,
            pierce: Optional[bool] = None
    ) -> Node:
        """
        Описывает узел с учетом его идентификатора, не требует включения домена. Не начинает отслеживать какие-либо
            объекты, можно использовать для автоматизации.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-describeNode
        :return:
        """
        if not ((nodeId != None) | (backendNodeId != None) | (objectId != None)):
            raise ValueError("Один из nodeId, backendNodeId, или objectId — должен присутствовать!")
        args = {}
        if nodeId is not None: args.update(nodeId=nodeId)
        if backendNodeId is not None: args.update(backendNodeId=backendNodeId)
        if objectId is not None: args.update(objectId=objectId)
        if depth is not None: args.update(depth=depth)
        if pierce is not None: args.update(pierce=pierce)
        result = await self._connection.call("DOM.describeNode", args)
        return Node(self._connection, **result["node"])

    async def resolveNode(
            self, nodeId: Optional[int] = None,
            backendNodeId: Optional[int] = None,
            objectGroup: Optional[str] = None,
            executionContextId: Optional[str] = None
    ) -> RemoteObject:
        """
        Создаёт JavaScript-объект для указанной ноды и возвращает его описание.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-resolveNode
        :return:
        """
        if not ((nodeId != None) | (backendNodeId != None)):
            raise ValueError("Один из nodeId, или backendNodeId — должен присутствовать!")
        args = {}
        if nodeId is not None: args.update(nodeId=nodeId)
        if backendNodeId is not None: args.update(backendNodeId=backendNodeId)
        if objectGroup is not None: args.update(objectGroup=objectGroup)
        if executionContextId is not None: args.update(executionContextId=executionContextId)
        result: dict = await self._connection.call("DOM.resolveNode", args)
        return RemoteObject(**result.get("object"))

    async def requestNode(self, objectId: str) -> Node:
        """
        Запрашивает, чтобы узел был отправлен вызывающей стороне с учетом ссылки на объект узла JavaScript.
            Все узлы, формирующие путь от узла к корню, также отправляются клиенту в виде серии
            setChildNodes-уведомлений.
        https://chromedevtools.github.io/devtools-protocol/tot/DOM/#method-requestNode
        :return:
        """
        args = {"objectId": objectId}
        result: dict = await self._connection.call("DOM.requestNode", args)
        return Node(self._connection, result.get("nodeId"))


class DOMEvent(DomainEvent):
    attributeModified = "DOM.attributeModified"
    attributeRemoved = "DOM.attributeRemoved"
    characterDataModified = "DOM.characterDataModified"
    childNodeCountUpdated = "DOM.childNodeCountUpdated"
    childNodeInserted = "DOM.childNodeInserted"
    childNodeRemoved = "DOM.childNodeRemoved"
    documentUpdated = "DOM.documentUpdated"
    setChildNodes = "DOM.setChildNodes"
    distributedNodesUpdated = "DOM.distributedNodesUpdated"     # ! EXPERIMENTAL
    inlineStyleInvalidated = "DOM.inlineStyleInvalidated"       # ! EXPERIMENTAL
    pseudoElementAdded = "DOM.pseudoElementAdded"               # ! EXPERIMENTAL
    pseudoElementRemoved = "DOM.pseudoElementRemoved"           # ! EXPERIMENTAL
    shadowRootPopped = "DOM.shadowRootPopped"                   # ! EXPERIMENTAL
    shadowRootPushed = "DOM.shadowRootPushed"                   # ! EXPERIMENTAL



from typing import Optional, List, Awaitable, Callable
from ...data import DomainEvent
from .types import TargetInfo


class Target:
    """
    #   https://chromedevtools.github.io/devtools-protocol/tot/Target
    """
    __slots__ = ("_connection", "targets_discovered")

    def __init__(self, conn) -> None:

        from ...connection import Connection

        self._connection: Connection = conn
        self.targets_discovered = False

    async def activateTarget(self, targetId: Optional[str] = None) -> None:
        """
        Активирует (создаёт фокус) "target".
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-activateTarget
        :param targetId:        Строка, представляющая идентификатор созданной страницы.
        :return:
        """
        if targetId is None: targetId = self._connection.conn_id
        await self._connection.call("Target.activateTarget", {"targetId": targetId})

    async def createBrowserContext(
        self,
        disposeOnDetach: Optional[bool] = None,
        proxyServer: Optional[str]      = None,
        proxyBypassList: Optional[str]  = None
    ) -> str:
        """
        Создает новый пустой BrowserContext. Аналогичен профилю в режиме инкогнито, но их может быть несколько.
        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-createBrowserContext
        :param disposeOnDetach:     Если True — удаляет контекст при отключении сеанса отладки.
        :param proxyServer:         Прокси-сервер, аналогичный тому, который передан --proxy-server.
        :param proxyBypassList:     Список обхода прокси, аналогичный тому, который передается в --proxy-bypass-list.
        :return:                Browser.BrowserContextID
        """
        args = {}
        if disposeOnDetach is not None: args.update(disposeOnDetach=disposeOnDetach)
        if proxyServer is not None: args.update(proxyServer=proxyServer)
        if proxyBypassList is not None: args.update(proxyBypassList=proxyBypassList)
        return (await self._connection.call("Target.createBrowserContext", args))["browserContextId"]

    async def getBrowserContexts(self) -> List[str]:
        """
        Возвращает все контексты браузера, созданные с помощью метода Target.createBrowserContext.
        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-getBrowserContexts
        :return:                [ Browser.BrowserContextID, Browser.BrowserContextID, ... ]
        """
        return (await self._connection.call("Target.getBrowserContexts"))["browserContextIds"]

    async def disposeBrowserContext(self, browserContextId: str) -> None:
        """
        Удаляет BrowserContext. Все соответствующие страницы будут закрыты без вызова их хуков beforeunload.
        https://chromedevtools.github.io/devtools-protocol/tot/Target/#method-disposeBrowserContext
        :param browserContextId:    Идентификатор контекста браузера.
        :return:
        """
        await self._connection.call("Target.disposeBrowserContext", {"browserContextId": browserContextId})

    async def getTargetInfo(self, targetId: Optional[str] = None) -> TargetInfo:
        """
        (EXPERIMENTAL)
        Возвращает информацию о "target", или о себе, если идентификатор не передан.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-getTargetInfo
        :param targetId:            Идентификатор страницы/сервиса/воркера/...
        :return:                    targetInfo -> {
                                        "targetId":         str,
                                        "type":             str,
                                        "title":            str,
                                        "url":              str,
                                        "attached":         bool,
                                        "openerId":         str,
                                        "browserContextId": str,
                                    }
        """
        if targetId is None: targetId = self._connection.conn_id
        return TargetInfo(
            **((await self._connection.call("Target.getTargetInfo", {"targetId": targetId}))["targetInfo"]))

    async def getTargets(self) -> List[TargetInfo]:
        """
        Возвращает список 'targetInfo' о доступных 'targets'.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-getTargets
        :return:                [ targetInfo, targetInfo, ... ]
        """
        result = (await self._connection.call("Target.getTargets"))["targetInfos"]
        return [TargetInfo(**info) for i, info in enumerate(result)]

    async def attachToBrowserTarget(self) -> str:
        """
        Присоединяется к target браузера, использует только режим flat sessionId.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-attachToBrowserTarget
        :return:                    sessionId
        """
        return (await self._connection.call("Target.attachToBrowserTarget"))["sessionId"]

    async def attachToTarget(self, targetId: str, flatten: Optional[bool] = None) -> str:
        """
        Присоединяется к 'target' по указанному 'targetId'.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-attachToTarget
        :param targetId:        Строка, представляющая идентификатор созданной страницы.
        :param flatten:         (optional) Разрешает "flat" доступ к сеансу с помощью указания атрибута
                                    sessionId в командах.
        :return:                sessionId -> Идентификатор, назначенный сеансу.
        """
        args = {"targetId": targetId}
        if flatten is not None: args.update({"flatten": flatten})
        return (await self._connection.call("Target.attachToTarget", args))["sessionId"]

    async def detachFromTarget(self, sessionId: str = "", targetId: str = "") -> None:
        """
        Отключается от сессии переданного 'sessionId'.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-detachFromTarget
        :param sessionId:       (optional) Строка, представляющая идентификатор сессии.
        :param targetId:        (optional) Строка, представляющая идентификатор созданной страницы.
        :return:
        """
        args = {}
        if sessionId: args.update({"sessionId": sessionId})
        elif targetId: args.update({"targetId": targetId})
        else: raise ValueError("At least one parameter must be specified 'sessionId' or 'targetId'")
        await self._connection.call("Target.detachFromTarget", args)

    async def setAutoAttach(
            self, autoAttach: bool, waitForDebuggerOnStart: bool, flatten: Optional[bool] = None) -> None:
        """
        Определяет, следует ли автоматически присоединяться к новым target, которые считаются связанными
            с этой. При включении также присоединяется ко всем существующим связанным целям. При
            выключении автоматически отсоединяется от всех присоединенных в данный момент целей. Это
            также удаляет все цели, добавленные autoAttachRelated из списка целей, чтобы отслеживать
            создание связанных целей.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-setAutoAttach
        :param autoAttach:                  Следует ли присоединяться к связанным targets.
        :param waitForDebuggerOnStart:      Приостанавливать ли новые targets при присоединении к ним.
                                                Используйте Runtime.runIfWaitingForDebugger для запуска
                                                приостановленных targets.
        :return:
        """
        args = {"autoAttach": autoAttach, "waitForDebuggerOnStart": waitForDebuggerOnStart}
        if flatten is not None: args.update(flatten=flatten)
        await self._connection.call("Target.setAutoAttach", args)

    async def createTarget(
        self,
        url: str                        = "about:blank",
        width: Optional[int]            = None,
        height: Optional[int]           = None,
        browserContextId: Optional[str] = None,
        enableBeginFrameControl: bool   = False,
        newWindow: bool                 = False,
        background: bool                = False
    ) -> str:
        """
        Создаёт новую страницу и возвращает её идентификатор. Чтобы затем управлять новой
            вкладкой, воспользуйтесь методом инстанса самого браузера GetPageByID(), или
            сразу методом CreatePage(), который проделывает все эти операции под капотом
            и возвращает готовый инстанс новой страницы.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-createTarget
        :param url:                         Адрес будет открыт при создании.
        :param width:                       Ширина в Density-independent Pixels(DIP). Chrome Headless only!
        :param height:                      Высота в (DIP). Chrome Headless only!
        :param browserContextId:            Контекст браузера, которому предназначается создание target.
        :param enableBeginFrameControl:     Будет ли BeginFrames для этой цели контролироваться с помощью DevTools
                                                (только безголовый хром, пока не поддерживается в MacOS, по
                                                умолчанию false).
        :param newWindow:                   Если 'True' — страница будет открыта в новом окне.
        :param background:                  Если 'True' — страница будет открыта в фоне.
        :return:                targetId -> строка, представляющая идентификатор созданной страницы.
        """
        args = {
            "url": url, "enableBeginFrameControl": enableBeginFrameControl,
            "newWindow": False if self._connection.is_headless_mode else newWindow,
            "background": False if self._connection.is_headless_mode else background
        }
        if width is not None: args.update(width=width)
        if height is not None: args.update(height=height)
        if browserContextId is not None: args.update(browserContextId=browserContextId)
        return (await self._connection.call("Target.createTarget", args))["targetId"]

    async def closeTarget(self, targetId: Optional[str] = None) -> None:
        """
        Закрывает вкладку указанного идентификатора, или завершает собственный инстанс,
            если идентификатор не передан.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-closeTarget
        :param targetId:        Строка, представляющая идентификатор созданной страницы.
        :return:                None
        """
        if targetId is None: targetId = self._connection.conn_id
        await self._connection.call("Target.closeTarget", {"targetId": targetId})

    async def close(self) -> None:
        """
        Закрывает вкладку.
        """
        await self.closeTarget()

    async def setDiscoverTargets(
        self, discover: bool,
        message:   Optional[Callable[[str, str], Awaitable[None]]] = None,
        created:   Optional[Callable[['TargetInfo'], Awaitable[None]]] = None,
        crashed:   Optional[Callable[[str, str, int], Awaitable[None]]] = None,
        changed:   Optional[Callable[['TargetInfo'], Awaitable[None]]] = None,
        destroyed: Optional[Callable[[str], Awaitable[None]]] = None
    ) -> None:
        """
        Управляет обнаружением доступных 'targets' уведомляя об их состоянии с помощью событий
            targetCreated / targetInfoChanged / targetDestroyed.
        https://chromedevtools.github.io/devtools-protocol/tot/Target#method-setDiscoverTargets
        :param discover:            'True' — включает эту надстройку, 'False' — выключает.
        :param message:             Корутина вызываемая для события 'Target.receivedMessageFromTarget'.
        :param created:             Корутина вызываемая для события 'Target.targetCreated'.
        :param crashed:             Корутина вызываемая для события 'Target.targetCrashed'.
        :param changed:             Корутина вызываемая для события 'Target.targetInfoChanged'.
        :param destroyed:           Корутина вызываемая для события 'Target.targetDestroyed'.
        :return:
        """
        async def on_message(params: dict) -> None:
            await message(params["sessionId"], params["message"])

        async def on_crashed(params: dict) -> None:
            await crashed(params["targetId"], params["status"], params["errorCode"])

        async def on_created(params: dict) -> None:
            await created(TargetInfo(**params["targetInfo"]))

        async def on_changed(params: dict) -> None:
            await changed(TargetInfo(**params["targetInfo"]))

        async def on_destroyed(params: dict) -> None:
            await destroyed(params["targetId"])

        if self.targets_discovered != discover:
            if discover:
                if message is not None:
                    await self._connection.addListenerForEvent(
                        TargetEvent.receivedMessageFromTarget, on_message)
                if created is not None:
                    await self._connection.addListenerForEvent(
                        TargetEvent.targetCreated, on_created)
                if crashed is not None:
                    await self._connection.addListenerForEvent(
                        TargetEvent.targetCrashed, on_crashed)
                if changed is not None:
                    await self._connection.addListenerForEvent(
                        TargetEvent.targetInfoChanged, on_changed)
                if destroyed is not None:
                    await self._connection.addListenerForEvent(
                        TargetEvent.targetDestroyed, on_destroyed)
            else:
                for event in [
                    'Target.receivedMessageFromTarget',
                    'Target.targetCreated',
                    'Target.targetCrashed',
                    'Target.targetInfoChanged',
                    'Target.targetDestroyed'
                ]:
                    self._connection.removeListenersForEvent(event)
            self.targets_discovered = discover
            await self._connection.call("Target.setDiscoverTargets", {"discover": discover})


class TargetEvent(DomainEvent):
    receivedMessageFromTarget = "Target.receivedMessageFromTarget"
    targetCrashed = "Target.targetCrashed"
    targetCreated = "Target.targetCreated"
    targetDestroyed = "Target.targetDestroyed"
    targetInfoChanged = "Target.targetInfoChanged"
    attachedToTarget = "Target.attachedToTarget"                        # ! EXPERIMENTAL
    detachedFromTarget = "Target.detachedFromTarget"                    # ! EXPERIMENTAL



from dataclasses import dataclass, field
from typing import Optional
from ..network.types import Request


@dataclass
class EventRequestPaused:
    requestId: str
    frameId: str
    resourceType: str                               # ! Allowed Values: Document, Stylesheet, Image, Media, Font,
                                                    # !     Script, TextTrack, XHR, Fetch, EventSource, WebSocket,
                                                    # !     Manifest, SignedExchange, Ping, CSPViolationReport,
                                                    # !     Preflight, Other
    request: Request
    responseHeaders: list["HeaderEntry"]
    responseErrorReason: Optional[str] = None       # ! Allowed Values: Failed, Aborted, TimedOut, AccessDenied,
                                                    # !     ConnectionClosed, ConnectionReset, ConnectionRefused,
                                                    # !     ConnectionAborted, ConnectionFailed, NameNotResolved,
                                                    # !     InternetDisconnected, AddressUnreachable,
                                                    # !     BlockedByClient, BlockedByResponse
    responseStatusCode: Optional[int] = None
    responseStatusText: Optional[str] = None
    networkId: Optional[str] = None                 # ! id of request
    redirectedRequestId: Optional[str] = None       # ! id запроса вызвавшего редирект
    _request: Request = field(init=False, repr=False, default=None)
    _responseHeaders: Optional[list["HeaderEntry"]] = field(init=False, repr=False, default=None)

    @property
    def request(self) -> Request:
        return self._request

    @request.setter
    def request(self, data: dict) -> None:
        self._request = Request(**data)

    @property
    def responseHeaders(self) -> Optional[list["HeaderEntry"]]:
        return self._responseHeaders

    @responseHeaders.setter
    def responseHeaders(self, data: list[dict[str, str]]) -> None:
        if not isinstance(data, property):
            self._responseHeaders = [HeaderEntry(**item) for item in data]
        else:
            self._responseHeaders = None


@dataclass
class EventAuthRequired:
    requestId: str
    request: Request
    frameId: str
    resourceType: str                               # ! Allowed Values: Document, Stylesheet, Image, Media, Font,
                                                    # !     Script, TextTrack, XHR, Fetch, EventSource, WebSocket,
                                                    # !     Manifest, SignedExchange, Ping, CSPViolationReport,
                                                    # !     Preflight, Other
    authChallenge: "AuthChallenge"
    _authChallenge: "AuthChallenge" = field(init=False, repr=False, default=None)

    @property
    def authChallenge(self) -> "AuthChallenge":
        return self._authChallenge

    @authChallenge.setter
    def authChallenge(self, data: dict) -> None:
        self._authChallenge = AuthChallenge(**data)


@dataclass
class AuthChallenge:
    origin: str
    scheme: str
    realm: str                                      # ! May be empty.
    source: Optional[str] = None


@dataclass
class HeaderEntry:
    name: str
    value: str


@dataclass
class RequestPattern:
    urlPattern: str = "*"                           # ? Подстановочные знаки ( '*'-> ноль или более, '?'-> ровно
                                                    # ?     один) допускаются. Экранирующий символ — обратная
                                                    # ?     косая черта(\).
    resourceType: Optional[str] = None              # ? Если установлено, будут перехватываться только соответствующие
                                                    # ?     указанным типам.
                                                    # ! Allowed Values: Document, Stylesheet, Image, Media, Font,
                                                    # !     Script, TextTrack, XHR, Fetch, EventSource, WebSocket,
                                                    # !     Manifest, SignedExchange, Ping, CSPViolationReport,
                                                    # !     Preflight, Other
    requestStage: Optional[str] = None              # ? Стадия перехвата запроса. Запрос будет перехвачен до того,
                                                    # ?     как запрос будет отправлен. Ответ будет перехвачен после
                                                    # ?     получения ответа (но до получения тела ответа).
                                                    # ! Allowed Values: Request, Response

    def dict(self) -> dict:
        result = {}
        for k, v in self.__dict__.items():
            if v: result[k] = v
        return result
